"""
Mixin for all implementations of :py:class:`RemoteAPI` for the Spotify API.

Also includes the default arguments to be used when requesting authorisation from the Spotify API.
"""
import base64
from collections.abc import Iterable
from copy import deepcopy

from musify import PROGRAM_NAME
from musify.api.exception import APIError
from musify.libraries.remote.spotify.api.item import SpotifyAPIItems
from musify.libraries.remote.spotify.api.misc import SpotifyAPIMisc
from musify.libraries.remote.spotify.api.playlist import SpotifyAPIPlaylists
from musify.libraries.remote.spotify.processors import SpotifyDataWrangler
from musify.utils import safe_format_map

URL_AUTH = "https://accounts.spotify.com"

# user authenticated access with scopes
SPOTIFY_API_AUTH_ARGS = {
    "auth_args": {
        "url": f"{URL_AUTH}/api/token",
        "data": {
            "grant_type": "authorization_code",
            "code": None,
            "redirect_uri": None,
        },
        "headers": {
            "content-type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {client_base64}"
        },
    },
    "user_args": {
        "url": f"{URL_AUTH}/authorize",
        "params": {
            "client_id": "{client_id}",
            "response_type": "code",
            "redirect_uri": None,
            "state": PROGRAM_NAME,
            "scope": "{scopes}",
            "show_dialog": False,
        },
    },
    "refresh_args": {
        "url": f"{URL_AUTH}/api/token",
        "data": {
            "grant_type": "refresh_token",
            "refresh_token": None,
        },
        "headers": {
            "content-type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {client_base64}"
        },
    },
    "test_args": {"url": "{url}/me"},
    "test_condition": lambda r: "href" in r and "display_name" in r,
    "test_expiry": 600,
    "token_key_path": ["access_token"],
    "header_extra": {"Accept": "application/json", "Content-Type": "application/json"},
}


class SpotifyAPI(SpotifyAPIMisc, SpotifyAPIItems, SpotifyAPIPlaylists):
    """
    Collection of endpoints for the Spotify API.

    :param client_id: The client ID to use when authorising requests.
    :param client_secret: The client secret to use when authorising requests.
    :param scopes: The scopes to request access to.
    """

    @property
    def user_id(self) -> str | None:
        """ID of the currently authenticated user"""
        if not self.user_data:
            try:
                self.user_data = self.get_self()
            except APIError:
                return None
        return self.user_data["id"]

    @property
    def user_name(self) -> str | None:
        """Name of the currently authenticated user"""
        if not self.user_data:
            try:
                self.user_data = self.get_self()
            except APIError:
                return None
        return self.user_data["display_name"]

    def __init__(
            self,
            client_id: str | None = None,
            client_secret: str | None = None,
            scopes: Iterable[str] = (),
            **kwargs,
    ):
        wrangler = SpotifyDataWrangler()
        format_map = {
            "client_id": client_id,
            "client_base64": base64.b64encode(f"{client_id}:{client_secret}".encode()).decode(),
            "scopes": " ".join(scopes),
            "url": wrangler.url_api
        }
        auth_kwargs = deepcopy(SPOTIFY_API_AUTH_ARGS)
        safe_format_map(auth_kwargs, format_map=format_map)

        super().__init__(wrangler=wrangler, **auth_kwargs, **kwargs)



"""
Implements all functionality pertaining to reading metadata/tags/properties for a :py:class:`LocalTrack`.
"""
import re
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from typing import Any

import mutagen
from PIL import Image

from musify.libraries.local.track.tags.base import TagProcessor
from musify.libraries.remote.core.enum import RemoteIDType
from musify.utils import to_collection


class TagReader[T: mutagen.FileType](TagProcessor, metaclass=ABCMeta):
    """Functionality for reading tags/metadata/properties from a mutagen object."""

    def read_tag(self, tag_ids: Iterable[str]) -> list[Any] | None:
        """Extract all tag values from file for a given list of tag IDs"""
        values = []
        for tag_id in tag_ids:
            value = self.file.get(tag_id)
            if value is None or (isinstance(value, str) and len(value.strip()) == 0):
                # skip null or empty/blank strings
                continue

            if isinstance(value, (list, set, tuple)) and all(isinstance(val, str) for val in value):
                values.extend(v for val in value for v in val.split('\x00'))
            elif isinstance(value, (list, set, tuple)):
                values.extend(value)
            elif isinstance(value, str):
                values.extend(value.split('\x00'))
            else:
                values.append(value)

        return values if len(values) > 0 else None

    def read_title(self) -> str | None:
        """Extract track title tags from file"""
        values = self.read_tag(self.tag_map.title)
        return str(values[0]) if values is not None else None

    def read_artist(self) -> str | None:
        """Extract artist tags from file"""
        values = self.read_tag(self.tag_map.artist)
        return str(values[0]) if values is not None else None

    def read_album(self) -> str | None:
        """Extract album tags from file"""
        values = self.read_tag(self.tag_map.album)
        return str(values[0]) if values is not None else None

    def read_album_artist(self) -> str | None:
        """Extract album artist tags from file"""
        values = self.read_tag(self.tag_map.album_artist)
        return str(values[0]) if values is not None else None

    def read_track_number(self) -> int | None:
        """Extract track number tags from file"""
        values = self.read_tag(self.tag_map.track_number)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[0]) if self.num_sep in value else int(value)

    def read_track_total(self) -> int | None:
        """Extract total track count tags from file"""
        values = self.read_tag(self.tag_map.track_total)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[1]) if self.num_sep in value else int(value)

    def read_genres(self) -> list[str] | None:
        """Extract genre tags from file"""
        values = self.read_tag(self.tag_map.genres)
        return list(map(str, values)) if values is not None else None

    def read_date(self) -> tuple[int | None, int | None, int | None] | None:
        """Extract year tags from file"""
        values = self.read_tag(self.tag_map.date)

        if values is None:  # attempt to read each part individually
            year = self.read_tag(self.tag_map.year)
            year = int(re.match(r"(\d{4})", str(year[0])).group(1)) if year else None
            month = self.read_tag(self.tag_map.month)
            month = int(re.match(r"(\d{1,2})", str(month[0])).group(1)) if month else None
            day = self.read_tag(self.tag_map.day)
            day = int(re.match(r"(\d{1,2})", str(day[0])).group(1)) if day else None
            return year, month, day
        elif 0 < len(values) <= 3 and all(str(value).isdigit() for value in values):
            values = ["/".join(values)]  # just join to string and allow later regex matches to determine format

        # YYYY-MM-DD
        match = re.match(r"(\d{4})\D+(\d{1,2})\D+(\d{1,2})", str(values[0]))
        if match:
            return int(match.group(1)), int(match.group(2)), int(match.group(3))

        # DD-MM-YY
        match = re.match(r"(\d{1,2})\D+(\d{1,2})\D+(\d{4})", str(values[0]))
        if match:
            return int(match.group(3)), int(match.group(2)), int(match.group(1))

        # YYYY-MM
        match = re.match(r"(\d{4})\D+(\d{1,2})", str(values[0]))
        if match:
            return int(match.group(1)), int(match.group(2)), None

        # MM-YYYY
        match = re.match(r"(\d{4})\D+(\d{1,2})", str(values[0]))
        if match:
            return int(match.group(2)), int(match.group(1)), None

        # YYYY
        match = re.match(r"(\d{4})", str(values[0]))
        if match:
            return int(match.group(1)), None, None

    def read_bpm(self) -> float | None:
        """Extract BPM tags from file"""
        values = self.read_tag(self.tag_map.bpm)
        return float(values[0]) if values is not None else None

    def read_key(self) -> str | None:
        """Extract key tags from file"""
        values = self.read_tag(self.tag_map.key)
        return str(values[0]) if values is not None else None

    def read_disc_number(self) -> int | None:
        """Extract disc number tags from file"""
        values = self.read_tag(self.tag_map.disc_number)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[0]) if self.num_sep in value else int(value)

    def read_disc_total(self) -> int | None:
        """Extract total disc count tags from file"""
        values = self.read_tag(self.tag_map.disc_total)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[1]) if self.num_sep in value else int(value)

    def read_compilation(self) -> bool | None:
        """Extract compilation tags from file"""
        values = self.read_tag(self.tag_map.compilation)
        return bool(int(values[0])) if values is not None else None

    def read_comments(self) -> list[str] | None:
        """Extract comment tags from file"""
        values = self.read_tag(self.tag_map.comments)
        return list({str(value) for value in values}) if values is not None else None

    def read_uri(self) -> str | None:
        """Extract data relating to remote URI value from file"""
        wrangler = self.remote_wrangler
        if not wrangler:
            return

        read_method = getattr(self, f"read_{self.uri_tag.name.lower()}")
        possible_values: tuple[str, ...] | None = to_collection(read_method())
        if not possible_values:
            return None

        # workaround for dodgy MP3 tag comments; split on null and take first value
        possible_values = tuple(val for values in possible_values for val in values.split("\x00"))
        for uri in possible_values:
            if uri == wrangler.unavailable_uri_dummy or wrangler.validate_id_type(uri, kind=RemoteIDType.URI):
                return uri

        return None

    @abstractmethod
    def read_images(self) -> list[Image.Image] | None:
        """Extract image from file"""
        raise NotImplementedError

    def check_for_images(self) -> bool:
        """Check if file has embedded images"""
        return self.read_tag(self.tag_map.images) is not None



"""
Processors that filter down objects and data types based on some given configuration.
"""
from __future__ import annotations

from collections.abc import Collection, Sequence, Mapping
from typing import Any, Self

from musify.core.base import MusifyObject
from musify.processors.base import Filter, FilterComposite
from musify.processors.compare import Comparer


class FilterDefinedList[T: str | MusifyObject](Filter[T], Collection[T]):

    __slots__ = ("values",)

    @property
    def ready(self):
        return len(self.values) > 0

    def __init__(self, values: Collection[T] = (), *_, **__):
        super().__init__()
        #: The values to include when processing for this filter
        self.values: Collection[T] = values

    def __call__(self, *args, **kwargs) -> Collection[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T] | None = None, *_, **__) -> Collection[T]:
        """Returns all ``values`` that match this filter's settings"""
        if self.ready:
            matches = [value for value in values if self.transform(value) in self.values]
            if isinstance(self.values, Sequence):
                matches = sorted((self.values.index(self.transform(match)), match) for match in matches)
                return [match[1] for match in matches]
            return matches
        return values

    def as_dict(self) -> dict[str, Any]:
        return {"values": self.values}

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __contains__(self, item: Any):
        return item in self.values


class FilterComparers[T: str | MusifyObject](Filter[T]):

    __slots__ = ("comparers", "match_all")

    @property
    def ready(self):
        return len(self.comparers) > 0

    def __init__(
            self,
            comparers: Collection[Comparer] | Mapping[Comparer, tuple[bool, Self]] = (),
            match_all: bool = True,
            *_,
            **__
    ):
        super().__init__()
        if not isinstance(comparers, Mapping):
            comparers = {comparer: (False, FilterComparers()) for comparer in comparers}

        #: The comparers to use when processing for this filter
        #: When a mapping with other :py:class:`FilterComparers` is given,
        #: will apply this sub-filter to all parent filters when processing
        self.comparers: Mapping[Comparer, tuple[bool, Self]] = comparers
        #: When true, only include those items that match on all comparers
        self.match_all: bool = match_all

    def __call__(self, *args, **kwargs) -> Collection[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T], reference: T | None = None, *_, **__) -> Collection[T]:
        if not self.ready:
            return values

        def run_comparer(c: Comparer, v: T) -> bool:
            """Run the comparer ``c`` for the given value ``v``"""
            return c(self.transform(v), reference=reference) if c.expected is None else c(self.transform(v))

        def run_sub_filter(s: FilterComparers, c: bool, v: list[T]) -> Collection[T]:
            """Run the sub_filter ``s`` for the given value ``v`` and return the results"""
            if not s.ready:
                return v
            elif c:
                return sub_filter(v, reference=reference)
            else:
                v.extend([value for value in sub_filter(values, reference=reference) if value not in v])
                return v

        if self.match_all:
            for comparer, (combine, sub_filter) in self.comparers.items():
                matched = [value for value in values if run_comparer(comparer, value)]
                values = run_sub_filter(sub_filter, combine, matched)
            return values

        matches = []
        for comparer, (combine, sub_filter) in self.comparers.items():
            matched = [value for value in values if value not in matches and run_comparer(comparer, value)]
            matched = run_sub_filter(sub_filter, combine, matched)
            matches.extend([value for value in matched if value not in matches])

        return matches

    def as_dict(self) -> dict[str, Any]:
        return {"comparers": self.comparers, "match_all": self.match_all}


###########################################################################
## Composites
###########################################################################
class FilterIncludeExclude[T: Any, U: Filter, V: Filter](FilterComposite[T]):

    __slots__ = ("include", "exclude")

    def __init__(self, include: U, exclude: V, *_, **__):
        super().__init__(include, exclude)
        #: The filter that, when processed, returns items to include
        self.include: U = include
        #: The filter that, when processed, returns items to exclude
        self.exclude: V = exclude

    def __call__(self, *args, **kwargs) -> list[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T], *_, **__) -> list[T]:
        """Filter down ``values`` that match this filter's settings from"""
        values = self.include.process(values) if self.include.ready else values
        exclude = self.exclude.process(values) if self.exclude.ready else ()
        return [v for v in values if v not in exclude]

    def as_dict(self) -> dict[str, Any]:
        return {"include": self.include, "exclude": self.exclude}



"""
The WMA implementation of a :py:class:`LocalTrack`.
"""
import struct
from collections.abc import Collection, Iterable
from io import BytesIO
from typing import Any

import mutagen
import mutagen.asf
import mutagen.id3
from PIL import Image, UnidentifiedImageError

from musify.core.enum import TagMap
from musify.file.image import open_image, get_image_bytes
from musify.libraries.local.track.tags.reader import TagReader
from musify.libraries.local.track.tags.writer import TagWriter
from musify.libraries.local.track.track import LocalTrack


class WMATagReader(TagReader[mutagen.asf.ASF]):

    def read_tag(self, tag_ids: Iterable[str]) -> list[Any] | None:
        # WMA tag values are returned as mutagen.asf._attrs.ASFUnicodeAttribute
        values = []
        for tag_id in tag_ids:
            value: Collection[mutagen.asf.ASFBaseAttribute] = self.file.get(tag_id)
            if value is None:
                # skip null or empty/blank strings
                continue

            values.extend([v.value for v in value if not (isinstance(value, str) and len(value.strip()) == 0)])

        return values if len(values) > 0 else None

    def read_images(self) -> list[Image.Image] | None:
        values = self.read_tag(self.tag_map.images)
        if values is None:
            return

        images: list[Image.Image] = []
        for i, value in enumerate(values):
            try:
                # first attempt to open image from bytes; assumes bytes value refer only to the image data
                images.append(Image.open(BytesIO(value)))
                continue
            except UnidentifiedImageError:
                # the bytes are encoded per WMA spec
                # bytes need to be analysed first to extract the bytes that refer to the image data
                pass

            v_type, v_size = struct.unpack_from(b"<bi", value)

            pos = 5
            mime = b""
            while value[pos:pos + 2] != b"\x00\x00":
                mime += value[pos:pos + 2]
                pos += 2

            pos += 2
            description = b""

            while value[pos:pos + 2] != b"\x00\x00":
                description += value[pos:pos + 2]
                pos += 2
            pos += 2

            image_data: bytes = value[pos:pos + v_size]
            images.append(Image.open(BytesIO(image_data)))

        return images


class WMATagWriter(TagWriter[mutagen.asf.ASF]):

    def write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool:
        result = super().write_tag(tag_id=tag_id, tag_value=tag_value, dry_run=dry_run)
        if result is not None:
            return result

        if not dry_run:
            if isinstance(tag_value, (list, set, tuple)):
                if all(isinstance(v, mutagen.asf.ASFByteArrayAttribute) for v in tag_value):
                    self.file[tag_id] = tag_value
                else:
                    self.file[tag_id] = [mutagen.asf.ASFUnicodeAttribute(str(v)) for v in tag_value]
            else:
                self.file[tag_id] = mutagen.asf.ASFUnicodeAttribute(str(tag_value))
        return True

    def _write_images(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id = next(iter(self.tag_map.images), None)

        updated = False
        tag_value = []
        for image_kind, image_link in track.image_links.items():
            image_kind_attr = image_kind.upper().replace(" ", "_")
            image_type: mutagen.id3.PictureType = getattr(mutagen.id3.PictureType, image_kind_attr)

            image = open_image(image_link)
            data = get_image_bytes(image)
            tag_data = struct.pack("<bi", image_type, len(data))
            tag_data += Image.MIME[image.format].encode("utf-16") + b"\x00\x00"  # mime
            tag_data += "".encode("utf-16") + b"\x00\x00"  # description
            tag_data += data

            tag_value.append(mutagen.asf.ASFByteArrayAttribute(tag_data))
            image.close()

        if len(tag_value) > 0:
            updated = self.write_tag(tag_id, tag_value, dry_run)

        track.has_image = updated or track.has_image
        return updated


class WMA(LocalTrack[mutagen.asf.ASF, WMATagReader, WMATagWriter]):

    valid_extensions = frozenset({".wma"})

    #: Map of human-friendly tag name to ID3 tag ids for a given file type
    tag_map = TagMap(
        title=["Title"],
        artist=["Author"],
        album=["WM/AlbumTitle"],
        album_artist=["WM/AlbumArtist"],
        track_number=["WM/TrackNumber"],
        track_total=["TotalTracks", "WM/TrackNumber"],
        genres=["WM/Genre"],
        year=["WM/Year", "WM/OriginalReleaseYear"],
        bpm=["WM/BeatsPerMinute"],
        key=["WM/InitialKey"],
        disc_number=["WM/PartOfSet"],
        disc_total=["WM/PartOfSet"],
        compilation=["COMPILATION"],
        comments=["Description", "WM/Comments"],
        images=["WM/Picture"],
    )

    @staticmethod
    def _create_reader(*args, **kwargs):
        return WMATagReader(*args, **kwargs)

    @staticmethod
    def _create_writer(*args, **kwargs):
        return WMATagWriter(*args, **kwargs)



"""
All classes and operations relating to the logger objects used throughout the entire package.
"""
import logging
import logging.config
import logging.handlers
import os
import sys
from collections.abc import Iterable
from typing import Any

from tqdm.auto import tqdm

from musify.log import INFO_EXTRA, REPORT, STAT


class MusifyLogger(logging.Logger):
    """The logger for all logging operations in Musify."""

    #: When true, never print a new line in the console when :py:meth:`print()` is called
    compact: bool = False
    #: When true, all bars returned by :py:meth:`get_progress_bar()` will be disabled by default
    disable_bars: bool = False
    #: All currently active progress bars
    _bars: list[tqdm] = []

    @property
    def file_paths(self) -> list[str]:
        """Get a list of the paths of all file handlers for this logger"""
        def extract_paths(lggr: logging.Logger) -> None:
            """Extract file path from the handlers of the given ``lggr``"""
            for handler in lggr.handlers:
                if isinstance(handler, logging.FileHandler) and handler.baseFilename not in paths:
                    paths.append(handler.baseFilename)

        paths = []
        logger = self
        extract_paths(logger)
        while logger.propagate and logger.parent:
            logger = logger.parent
            extract_paths(logger)
        return paths

    @property
    def stdout_handlers(self) -> list[logging.StreamHandler]:
        """Get a list of all :py:class:`logging.StreamHandler` handlers that log to stdout"""
        console_handlers = []
        for handler in self.handlers:
            if isinstance(handler, logging.StreamHandler) and handler.stream == sys.stdout:
                console_handlers.append(handler)
        return console_handlers

    def __init__(self, name: str, level: int | str = logging.NOTSET):
        super().__init__(name=name, level=level)

    def info_extra(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'INFO_EXTRA'."""
        if self.isEnabledFor(INFO_EXTRA):
            self._log(INFO_EXTRA, msg, args, **kwargs)

    def report(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'REPORT'."""
        if self.isEnabledFor(REPORT):
            self._log(REPORT, msg, args, **kwargs)

    def stat(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'STAT'."""
        if self.isEnabledFor(STAT):
            self._log(STAT, msg, args, **kwargs)

    def print(self, level: int = logging.CRITICAL + 1) -> None:
        """Print a new line only when DEBUG < ``logger level`` <= ``level`` for all console handlers"""
        if not self.compact and all(logging.DEBUG < h.level <= level for h in self.stdout_handlers):
            print()

    def get_progress_bar[T: Any](
            self,
            iterable: Iterable[T] | None = None,
            total: T | int | None = None,
            **kwargs
    ) -> tqdm | Iterable[T]:
        """Wrapper for tqdm progress bar. For kwargs, see :py:class:`tqdm_std`"""
        # noinspection SpellCheckingInspection
        preset_keys = ("leave", "disable", "file", "ncols", "colour", "smoothing", "position")

        try:
            cols = os.get_terminal_size().columns
        except OSError:
            cols = 120

        # clear closed bars
        self._bars = [bar for bar in self._bars if bar.n < bar.total]

        # determine the level of bar to generate and whether to leave the bar based on current active count
        position = kwargs.get("position", abs(min(bar.pos for bar in self._bars)) + 1 if self._bars else 0)
        leave_default = all(h.level > logging.DEBUG for h in self.stdout_handlers) and position == 0
        leave = kwargs["leave"] if kwargs.get("leave") is not None else leave_default

        bar = tqdm(
            iterable=iterable,
            total=total,
            leave=leave,
            disable=self.disable_bars or kwargs.get("disable", False),
            file=sys.stdout,
            ncols=cols,
            colour=kwargs.get("colour", "blue"),
            smoothing=0.1,
            position=position,
            **{k: v for k, v in kwargs.items() if k not in preset_keys}
        )
        self._bars.append(bar)
        return bar

    def __copy__(self):
        """Do not copy logger"""
        return self

    def __deepcopy__(self, _: dict = None):
        """Do not copy logger"""
        return self


logging.setLoggerClass(MusifyLogger)



"""
All type hints to use throughout the module.
"""
from collections.abc import MutableMapping
from typing import Any, TypeVar

from musify.libraries.remote.core import RemoteResponse
from musify.types import UnitMutableSequence, UnitSequence

UT = TypeVar('UT')
APIInputValue = (
        UnitMutableSequence[str] |
        UnitMutableSequence[MutableMapping[str, Any]] |
        UnitSequence[RemoteResponse]
)



"""
All core type hints to use throughout the entire package.
"""
from collections.abc import Iterable, Sequence, MutableSequence, Collection, Mapping, MutableMapping
from typing import TypeVar

UT = TypeVar('UT')
UnitIterable = UT | Iterable[UT]
UnitCollection = UT | Collection[UT]
UnitSequence = UT | Sequence[UT]
UnitMutableSequence = UT | MutableSequence[UT]
UnitList = UT | list[UT]

JSON_VALUE = str | int | float | list | dict | bool | None
JSON = Mapping[str, JSON_VALUE]
MutableJSON = MutableMapping[str, JSON_VALUE]
DictJSON = dict[str, JSON_VALUE]

Number = int | float



"""
All classes and functions pertaining to logging operations throughout the package.
"""
import logging

LOGGING_DT_FORMAT = "%Y-%m-%d_%H.%M.%S"

INFO_EXTRA = logging.INFO - 1
logging.addLevelName(INFO_EXTRA, "INFO_EXTRA")
logging.INFO_EXTRA = INFO_EXTRA

REPORT = logging.INFO - 3
logging.addLevelName(REPORT, "REPORT")
logging.REPORT = REPORT

STAT = logging.DEBUG + 3
logging.addLevelName(STAT, "STAT")
logging.STAT = STAT



"""
The fundamental core result classes for the entire package.
"""
from abc import ABC
from dataclasses import dataclass


@dataclass(frozen=True)
class Result(ABC):
    """Stores the results of an operation"""
    pass



"""
The fundamental core enum classes for the :py:mod:`Remote` module.

Represents ID and item types.
"""
from musify.core.enum import MusifyEnum


class RemoteIDType(MusifyEnum):
    """Represents remote ID types"""
    ALL: int = 0

    #: Value is the expected length of ID string
    ID: int = 22
    #: Value is the expected number of chunks in a URI
    URI: int = 3
    URL: int = 1
    URL_EXT: int = 2


class RemoteObjectType(MusifyEnum):
    """Represents remote object types"""
    ALL = 0
    PLAYLIST = 1
    TRACK = 2
    ALBUM = 3
    ARTIST = 4
    USER = 5
    SHOW = 6
    EPISODE = 7
    AUDIOBOOK = 8
    CHAPTER = 9



"""
Implements endpoints for getting items from the Spotify API.
"""
import re
from abc import ABCMeta
from collections.abc import Collection, Mapping, MutableMapping
from itertools import batched
from typing import Any
from urllib.parse import parse_qs, urlparse

from musify.api.exception import APIError
from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.enum import RemoteIDType, RemoteObjectType
from musify.libraries.remote.core.exception import RemoteObjectTypeError
from musify.libraries.remote.core.types import APIInputValue
from musify.libraries.remote.spotify.api.base import SpotifyAPIBase
from musify.utils import limit_value

ARTIST_ALBUM_TYPES = {"album", "single", "compilation", "appears_on"}


class SpotifyAPIItems(SpotifyAPIBase, metaclass=ABCMeta):

    _bar_threshold = 5

    def _get_unit(self, key: str | None = None, kind: str | None = None) -> str:
        """Determine the unit type to use in the progress bar"""
        if kind is None:
            kind = re.sub(r"[-_]+", " ", key) if key is not None else self.items_key
        return kind.casefold().rstrip("s") + "s"

    ###########################################################################
    ## GET helpers: Generic methods for getting items
    ###########################################################################
    def _get_items_multi(
            self,
            url: str,
            id_list: Collection[str],
            params: Mapping[str, Any] | None = None,
            key: str | None = None,
            kind: str | None = None,
            use_cache: bool = True,
            *_,
            **__,
    ) -> list[dict[str, Any]]:
        """
        Get responses from a given ``url`` appending an ID to each request from a given ``id_list``
        i.e. ``URL`` or ``ID``.
        This function executes each URL request individually for each ID i.e. ``URL`` or ``ID``.

        :param url: The base API URL endpoint for the required requests.
        :param id_list: List of IDs to append to the given URL.
        :param params: Extra parameters to add to each request.
        :param kind: The unit to use for logging. If None, determine from ``key`` or default to ``items``.
        :param key: The key to reference from each response to get the list of required values.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each item at the given ``key``.
        :raise APIError: When the given ``key`` is not in the API response.
        """
        url = url.rstrip("/")
        kind = self._get_unit(key=key, kind=kind)

        bar = self.logger.get_progress_bar(
            iterable=id_list, desc=f"Getting {kind}", unit=kind, disable=len(id_list) < self._bar_threshold
        )

        results: list[dict[str, Any]] = []
        log = [f"{kind.title()}:{len(id_list):>5}"]
        for id_ in bar:
            response = self.handler.get(f"{url}/{id_}", params=params, use_cache=use_cache, log_pad=43, log_extra=log)
            if "id" not in response:
                response["id"] = id_
            if key and key not in response:
                raise APIError(f"Given key '{key}' not found in response keys: {list(response.keys())}")

            results.extend(response[key]) if key else results.append(response)

        return results

    def _get_items_batched(
            self,
            url: str,
            id_list: Collection[str],
            params: Mapping[str, Any] | None = None,
            key: str | None = None,
            kind: str | None = None,
            limit: int = 50,
            use_cache: bool = True,
            *_,
            **__,
    ) -> list[dict[str, Any]]:
        """
        Get responses from a given ``url`` appending an ID to each request from a given ``id_list``
        i.e. ``URL`` or ``ID``.
        This function executes each URL request in batches of IDs based on the given ``batch_size``.
        It passes this chunked list of IDs to the request handler as a set of params in the form:
        ``{<IDs>: '<comma separated string of IDs>'}``

        :param url: The base API URL endpoint for the required requests.
        :param id_list: List of IDs to append to the given URL.
        :param params: Extra parameters to add to each request.
        :param kind: The unit to use for logging. If None, determine from ``key`` or default to ``items``.
        :param key: The key to reference from each response to get the list of required values.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :param limit: Size of each batch of IDs to get. This value will be limited to be between ``1`` and ``50``.
        :return: API JSON responses for each item at the given ``key``.
        :raise APIError: When the given ``key`` is not in the API response.
        """
        url = url.rstrip("/")
        kind = self._get_unit(key=key, kind=kind)

        id_chunks = list(batched(id_list, limit_value(limit, floor=1, ceil=50)))
        bar = self.logger.get_progress_bar(
            iterable=range(len(id_chunks)),
            desc=f"Getting {kind}",
            unit="pages",
            disable=len(id_chunks) < self._bar_threshold
        )

        results: list[dict[str, Any]] = []
        params = params if params is not None else {}
        for idx in bar:  # get responses in batches
            id_chunk = id_chunks[idx]
            params_chunk = params | {"ids": ",".join(id_chunk)}
            log = [f"{kind.title() + ':':<11} {len(results) + len(id_chunk):>6}/{len(id_list):<6}"]

            response = self.handler.get(url, params=params_chunk, use_cache=use_cache, log_pad=43, log_extra=log)
            if key and key not in response:
                raise APIError(f"Given key '{key}' not found in response keys: {list(response.keys())}")

            results.extend(response[key]) if key else results.append(response)

        return results

    ###########################################################################
    ## GET endpoints
    ###########################################################################
    def extend_items(
            self,
            response: MutableMapping[str, Any] | RemoteResponse,
            kind: RemoteObjectType | str | None = None,
            key: RemoteObjectType | None = None,
            use_cache: bool = True,
            leave_bar: bool | None = None,
    ) -> list[dict[str, Any]]:
        """
        Extend the items for a given API ``response``.
        The function requests each page of the collection returning a list of all items
        found across all pages for this URL.

        Updates the value of the ``items`` key in-place by extending the value of the ``items`` key with new results.

        If a :py:class:`RemoteResponse`, this function will not refresh itself with the new response.
        The user must call `refresh` manually after execution.

        :param response: A remote API JSON response for an items type endpoint
            or a RemoteResponse which contains this response.
            Must include required keys:
            ``total`` and either ``next`` or ``href``, plus optional keys ``previous``, ``limit``, ``items`` etc.
        :param kind: The type of response being extended. Optional, used only for logging.
        :param key: The type of response of the child objects. Used when selecting nested data for certain responses
            (e.g. user's followed artists).
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :param leave_bar: When a progress bar is displayed,
            toggle whether this bar should continue to be displayed after the operation is finished.
            When None, allow the logger to decide this setting.
        :return: API JSON responses for each item
        """
        if isinstance(response, RemoteResponse):
            response = response.response

        key = key.name.lower() + "s" if key else None
        if key and key in response:
            response = response[key]
        if self.items_key not in response:
            response[self.items_key] = []

        if len(response[self.items_key]) == response["total"]:  # skip on fully extended response
            url = response["href"].split("?")[0]
            self.logger.debug(f"{'SKIP':<7}: {url:<43} | Response already extended")
            return response[self.items_key]

        # this usually happens on the items block of a current user's playlist
        if "next" not in response:
            response["next"] = response["href"]
        if "previous" not in response:
            response["previous"] = None
        if "limit" not in response:
            response["limit"] = int(parse_qs(urlparse(response["next"]).query).get("limit", [50])[0])

        kind = kind.name.lower() + "s" if isinstance(kind, RemoteObjectType) else kind or self.items_key
        pages = (response["total"] - len(response[self.items_key])) / (response["limit"] or 1)
        bar = self.logger.get_progress_bar(
            initial=len(response[self.items_key]),
            total=response["total"],
            desc=f"Extending {kind}".rstrip("s") if kind[0].islower() else kind,
            unit=key or self.items_key,
            leave=leave_bar,
            disable=pages < self._bar_threshold,
        )

        while response.get("next"):  # loop through each page
            log_count = min(len(response[self.items_key]) + response["limit"], response["total"])
            log = [f"{log_count:>6}/{response["total"]:<6} {key or self.items_key}"]

            response_next = self.handler.get(response["next"], use_cache=use_cache, log_pad=95, log_extra=log)
            if key and key in response_next:
                response_next = response_next[key]

            response[self.items_key].extend(response_next[self.items_key])
            response["href"] = response_next["href"]
            response["next"] = response_next.get("next")
            response["previous"] = response_next.get("previous")

            bar.update(len(response_next[self.items_key]))

        bar.close()
        return response[self.items_key]

    def get_items(
            self,
            values: APIInputValue,
            kind: RemoteObjectType | None = None,
            limit: int = 50,
            extend: bool = True,
            use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /{kind}s`` - Get information for given ``values``.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection including:
                - some items under an ``items`` key,
                - a valid ID value under an ``id`` key,
                - a valid item type value under a ``type`` key if ``kind`` is None.
            * A MutableSequence of remote API JSON responses for a collection including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this update each response given by merging with the new response
        and replacing the ``items`` with the new results.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items must all be of the same type of item i.e. all tracks OR all artists etc.
        :param kind: Item type if given string is ID.
        :param limit: When requests can be batched, size of batches to request.
            This value will be limited to be between ``1`` and ``50`` or ``20`` if getting albums.
        :param extend: When True and the given ``kind`` is a collection of items,
            extend the response to include all items in this collection.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each item.
        :raise RemoteObjectTypeError: Raised when the function cannot determine the item type
            of the input ``values``. Or when it does not recognise the type of the input ``values`` parameter.
        """
        # input validation
        if not isinstance(values, RemoteResponse) and not values:  # skip on empty
            url = f"{self.url}/{kind.name.lower() + "s"}" if kind else self.url
            self.logger.debug(f"{'SKIP':<7}: {url:<43} | No data given")
            return []
        if kind is None:  # determine the item type
            kind = self.wrangler.get_item_type(values)
        else:
            self.wrangler.validate_item_type(values, kind=kind)

        unit = kind.name.lower() + "s"
        url = f"{self.url}/{unit}"
        id_list = self.wrangler.extract_ids(values, kind=kind)

        if kind in {RemoteObjectType.USER, RemoteObjectType.PLAYLIST} or len(id_list) <= 1:
            results = self._get_items_multi(url=url, id_list=id_list, kind=unit, use_cache=use_cache)
        else:
            if kind == RemoteObjectType.ALBUM:
                limit = limit_value(limit, floor=1, ceil=20)
            results = self._get_items_batched(url=url, id_list=id_list, key=unit, use_cache=use_cache, limit=limit)

        key = self.collection_item_map.get(kind, kind)
        key_name = key.name.lower() + "s"
        if len(results) == 0 or any(key_name not in result for result in results) or not extend:
            self._merge_results_to_input(original=values, responses=results, ordered=True)
            self._refresh_responses(responses=values, skip_checks=False)
            self.logger.debug(f"{'DONE':<7}: {url:<43} | Retrieved {len(results):>6} {unit}")
            return results

        bar = self.logger.get_progress_bar(
            iterable=results, desc=f"Extending {unit}", unit=unit, disable=len(id_list) < self._bar_threshold
        )

        for result in bar:
            if result[key_name].get("next") or ("next" not in result[key_name] and result[key_name].get("href")):
                self.extend_items(result[key_name], kind=kind, key=key, use_cache=use_cache, leave_bar=False)

        self._merge_results_to_input(original=values, responses=results, ordered=True)
        self._refresh_responses(responses=values, skip_checks=False)

        item_count = sum(len(result[key_name][self.items_key]) for result in results)
        self.logger.debug(
            f"{'DONE':<7}: {url:<71} | Retrieved {item_count:>6} {key_name} across {len(results):>5} {unit}"
        )

        return results

    def get_user_items(
            self,
            user: str | None = None,
            kind: RemoteObjectType = RemoteObjectType.PLAYLIST,
            limit: int = 50,
            use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /{kind}s`` - Get saved items for a given user.

        :param user: The ID of the user to get playlists for. If None, use the currently authenticated user.
        :param kind: Item type to retrieve for the user.
            Spotify only supports ``PLAYLIST`` types for non-authenticated users.
        :param limit: Size of each batch of items to request in the collection items request.
            This value will be limited to be between ``1`` and ``50``.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each collection.
        :raise RemoteIDTypeError: Raised when the input ``user`` does not represent a user URL/URI/ID.
        :raise RemoteObjectTypeError: Raised a user is given and the ``kind`` is not ``PLAYLIST``.
            Or when the given ``kind`` is not a valid collection.
        """
        # input validation
        if kind not in self.user_item_types:
            raise RemoteObjectTypeError(f"{kind.name.title()}s are not a valid user collection type", kind=kind)
        if kind != RemoteObjectType.PLAYLIST and user is not None:
            raise RemoteObjectTypeError(
                f"Only able to retrieve {kind.name.lower()}s from the currently authenticated user",
                kind=kind
            )

        params = {"limit": limit_value(limit, floor=1, ceil=50)}
        if user is not None:
            url = self.wrangler.convert(user, kind=RemoteObjectType.USER, type_out=RemoteIDType.URL)
            url = f"{url}/{kind.name.lower()}s"
            desc_qualifier = "user's"
        elif kind == RemoteObjectType.ARTIST:
            url = f"{self.url}/me/following"
            desc_qualifier = "current user's followed"
            params["type"] = "artist"
        else:
            url = f"{self.url}/me/{kind.name.lower()}s"
            desc_qualifier = "current user's" if kind == RemoteObjectType.PLAYLIST else "current user's saved"

        desc = f"Getting {desc_qualifier} {kind.name.lower()}s"
        initial = self.handler.get(url, params=params, use_cache=use_cache, log_pad=71)
        results = self.extend_items(initial, kind=desc, key=kind, use_cache=use_cache)

        self.logger.debug(f"{'DONE':<7}: {url:<43} | Retrieved {len(results):>6} {kind.name.lower()}s")

        return results

    def extend_tracks(
            self,
            values: APIInputValue,
            features: bool = False,
            analysis: bool = False,
            limit: int = 50,
            use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /audio-features`` and/or ``GET: /audio-analysis`` - Get audio features/analysis for given track/s.

        ``values`` may be:
            * A string representing a URL/URI/ID of type 'track'.
            * A MutableSequence of strings representing URLs/URIs/IDs of the type 'track'.
            * A remote API JSON response for a track including:
                - some items under an ``items`` key,
                - a valid ID value under an ``id`` key.
            * A MutableSequence of remote API JSON responses for a set of tracks including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by adding the results
        under the ``audio_features`` and ``audio_analysis`` keys as appropriate.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote track/s. See description for allowed value types.
        :param features: When True, get audio features.
        :param analysis: When True, get audio analysis.
        :param limit: Size of batches to request when getting audio features.
            This value will be limited to be between ``1`` and ``50``.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each item.
            Mapped to ``audio_features`` and ``audio_analysis`` keys as appropriate.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``values`` are not all tracks or IDs.
        """
        # input validation
        if not features and not analysis:  # skip on all False
            return []
        if not values:  # skip on empty
            self.logger.debug(f"{'SKIP':<7}: {self.url:<43} | No data given")
            return []
        self.wrangler.validate_item_type(values, kind=RemoteObjectType.TRACK)

        id_list = self.wrangler.extract_ids(values, kind=RemoteObjectType.TRACK)

        # value list takes the form [URL, key, batched]
        config: dict[str, tuple[str, str, bool]] = {}
        if features:
            config["features"] = (f"{self.url}/audio-features", "audio_features", True)
        if analysis:
            config["analysis"] = (f"{self.url}/audio-analysis", "audio_analysis", False)

        results: list[dict[str, Any]]
        if len(id_list) == 1:
            id_ = id_list[0]
            id_map = {self.id_key: id_}

            result = id_map.copy()
            for (url, key, _) in config.values():
                result[key] = self.handler.get(f"{url}/{id_}", use_cache=use_cache, log_pad=43) | id_map.copy()
            results = [result]
        else:
            results = []
            for kind, (url, key, batch) in config.items():
                method = self._get_items_batched if batch else self._get_items_multi
                responses = method(
                    url=url, id_list=id_list, kind=kind, key=key if batch else None, limit=limit, use_cache=use_cache
                )
                responses.sort(key=lambda response: id_list.index(response[self.id_key]))
                responses = [{self.id_key: response[self.id_key], key: response} for response in responses]

                if not results:
                    results = responses
                else:
                    results = [result | response for result, response in zip(results, responses)]

        self._merge_results_to_input(original=values, responses=results, ordered=False, clear=False)
        self._refresh_responses(responses=values, skip_checks=False)

        def map_key(value: str) -> str:
            """Map the given ``value`` to logging appropriate string"""
            return value.replace("_", " ").replace("analysis", "analyses")

        log_url = f"{self.url}/{"+".join(c[0].split("/")[-1] for c in config.values())}"
        self.logger.debug(
            f"{'DONE':<7}: {log_url:<71} | "
            f"Retrieved {" and ".join(map_key(key) for _, key, _ in config.values())} for {len(id_list):>5} tracks"
        )

        return results

    def get_tracks(
            self,
            values: APIInputValue,
            features: bool = False,
            analysis: bool = False,
            limit: int = 50,
            use_cache: bool = True,
            *_,
            **__,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /{kind}s`` + ``GET: /audio-features`` and/or ``GET: /audio-analysis``

        Get track(s) info and any audio features/analysis.
        Mostly just a wrapper for ``get_items`` and ``extend_tracks`` functions.

        ``values`` may be:
            * A string representing a URL/URI/ID of type 'track'.
            * A MutableSequence of strings representing URLs/URIs/IDs of the type 'track'.
            * A remote API JSON response for a track including:
                - some items under an ``items`` key,
                - a valid ID value under an ``id`` key.
            * A MutableSequence of remote API JSON responses for a set of tracks including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by adding the results
        under the ``audio_features`` and ``audio_analysis`` keys as appropriate.

        :param values: The values representing some remote track/s. See description for allowed value types.
        :param features: When True, get audio features.
        :param analysis: When True, get audio analysis.
        :param limit: Size of batches to request when getting audio features.
            This value will be limited to be between ``1`` and ``50``.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each item, or the original response if the input ``values`` are API responses.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``values`` are not all tracks or IDs.
        """
        tracks = self.get_items(values=values, kind=RemoteObjectType.TRACK, limit=limit, use_cache=use_cache)

        # ensure that response are being assigned back to the original values if API response(s) given
        if isinstance(values, Mapping | RemoteResponse):
            tracks = [values]
        elif isinstance(values, Collection) and all(isinstance(v, Mapping | RemoteResponse) for v in values):
            tracks = values

        self.extend_tracks(values=tracks, features=features, analysis=analysis, limit=limit, use_cache=use_cache)
        return tracks

    def get_artist_albums(
            self,
            values: APIInputValue,
            types: Collection[str] = (),
            limit: int = 50,
            use_cache: bool = True,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        ``GET: /artists/{ID}/albums`` - Get all albums associated with the given artist/s.

        ``values`` may be:
            * A string representing a URL/URI/ID of type 'artist'.
            * A MutableSequence of strings representing URLs/URIs/IDs of the type 'artist'.
            * A remote API JSON response for an artist including a valid ID value under an ``id`` key.
            * A MutableSequence of remote API JSON responses for a set of artists including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by adding the results under the ``albums`` key.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote artist/s. See description for allowed value types.
        :param types: The types of albums to return. Select from ``{"album", "single", "compilation", "appears_on"}``.
        :param limit: Size of batches to request.
            This value will be limited to be between ``1`` and ``50``.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: A map of the Artist ID to a list of the API JSON responses for each album.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``values`` are not all artists or IDs.
        """
        url = f"{self.url}/artists/{{id}}/albums"

        # input validation
        if not isinstance(values, RemoteResponse) and not values:  # skip on empty
            self.logger.debug(f"{'SKIP':<7}: {url:<43} | No data given")
            return {}

        if types and not all(t in ARTIST_ALBUM_TYPES for t in types):
            raise APIError(
                f"Given types not recognised, must be one or many of the following: {ARTIST_ALBUM_TYPES} ({types})"
            )
        self.wrangler.validate_item_type(values, kind=RemoteObjectType.ARTIST)

        id_list = self.wrangler.extract_ids(values, kind=RemoteObjectType.ARTIST)
        bar = self.logger.get_progress_bar(
            iterable=id_list, desc="Getting artist albums", unit="artist", disable=len(id_list) < self._bar_threshold
        )

        params = {"limit": limit_value(limit, floor=1, ceil=50)}
        if types:
            params["include_groups"] = ",".join(set(types))

        key = RemoteObjectType.ALBUM
        results: dict[str, dict[str, Any]] = {}
        for id_ in bar:
            results[id_] = self.handler.get(url=url.format(id=id_), params=params, use_cache=use_cache)
            self.extend_items(results[id_], kind="artist albums", key=key, use_cache=use_cache, leave_bar=False)

            for album in results[id_]["items"]:  # add skeleton items block to album responses
                album["tracks"] = {
                    "href": self.format_next_url(url=album["href"].split("?")[0] + "/tracks", offset=0, limit=50),
                    "total": album["total_tracks"]
                }

        results_remapped = [{"id": id_, "albums": result} for id_, result in results.items()]
        self._merge_results_to_input(original=values, responses=results_remapped, ordered=False, clear=False)
        self._refresh_responses(responses=values, skip_checks=True)

        item_count = sum(len(result) for result in results.values())
        self.logger.debug(
            f"{'DONE':<7}: {url.format(id="..."):<71} | "
            f"Retrieved {item_count:>6} albums across {len(results):>5} artists"
        )

        return {k: v["items"] for k, v in results.items()}



"""
All core :py:class:`Field` implementations relating to
core :py:class:`MusifyItem` and :py:class`MusifyCollection` implementations.
"""
from typing import Self

from musify.core.enum import Field, Fields, TagField, TagFields


class TrackFieldMixin(TagField):
    """Applies extra functionality to the TagField enum for TagField types relating to :py:class:`Track` types"""

    # noinspection PyUnresolvedReferences
    @classmethod
    def map(cls, enum: Self) -> list[Self]:
        if enum == cls.TRACK:
            return [cls.TRACK_NUMBER, cls.TRACK_TOTAL]
        elif enum == cls.DISC:
            return [cls.DISC_NUMBER, cls.DISC_TOTAL]
        return [enum]


class TrackField(TrackFieldMixin):
    """Represents all currently supported fields for objects of type :py:class:`Track`"""
    ALL = TagFields.ALL.value

    TITLE = TagFields.TITLE.value
    ARTIST = TagFields.ARTIST.value
    ALBUM = TagFields.ALBUM.value
    ALBUM_ARTIST = TagFields.ALBUM_ARTIST.value
    TRACK = TagFields.TRACK_NUMBER.value + 500
    TRACK_NUMBER = TagFields.TRACK_NUMBER.value
    TRACK_TOTAL = TagFields.TRACK_TOTAL.value
    GENRES = TagFields.GENRES.value
    YEAR = TagFields.YEAR.value
    BPM = TagFields.BPM.value
    KEY = TagFields.KEY.value
    DISC = TagFields.DISC_NUMBER.value + 500
    DISC_NUMBER = TagFields.DISC_NUMBER.value
    DISC_TOTAL = TagFields.DISC_TOTAL.value
    COMPILATION = TagFields.COMPILATION.value
    COMMENTS = TagFields.COMMENTS.value
    IMAGES = TagFields.IMAGES.value
    LENGTH = TagFields.LENGTH.value
    RATING = TagFields.RATING.value

    # remote properties
    URI = TagFields.URI.value


class PlaylistField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Playlist`"""
    ALL = Fields.ALL.value

    # tags/core properties
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value

    # date properties
    DATE_CREATED = Fields.DATE_CREATED.value
    DATE_MODIFIED = Fields.DATE_MODIFIED.value

    # miscellaneous properties
    DESCRIPTION = Fields.DESCRIPTION.value


class FolderField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Folder`"""
    ALL = Fields.ALL.value

    # tags/core properties
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    IMAGES = Fields.IMAGES.value
    COMPILATION = Fields.COMPILATION.value
    LENGTH = Fields.LENGTH.value

    # file properties
    FOLDER = Fields.FOLDER.value


class AlbumField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Album`"""
    ALL = Fields.ALL.value

    # tags/core properties
    ARTIST = Fields.ARTIST.value
    ALBUM = Fields.ALBUM.value
    ALBUM_ARTIST = Fields.ALBUM_ARTIST.value
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    DATE = TagFields.DATE.value
    YEAR = TagFields.YEAR.value
    MONTH = TagFields.MONTH.value
    DAY = TagFields.DAY.value
    DISC_TOTAL = Fields.DISC_TOTAL.value
    COMPILATION = Fields.COMPILATION.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value
    RATING = Fields.RATING.value


class ArtistField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Artist`"""
    ALL = Fields.ALL.value

    # tags/core properties
    ARTIST = Fields.ARTIST.value
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value
    RATING = Fields.RATING.value



"""
Implements endpoints for manipulating playlists with the Spotify API.
"""
from abc import ABCMeta
from collections.abc import Collection, Mapping
from itertools import batched
from typing import Any

from musify import PROGRAM_NAME, PROGRAM_URL
from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.enum import RemoteIDType, RemoteObjectType
from musify.libraries.remote.core.exception import RemoteIDTypeError
from musify.libraries.remote.spotify.api.base import SpotifyAPIBase
from musify.utils import limit_value


class SpotifyAPIPlaylists(SpotifyAPIBase, metaclass=ABCMeta):
    """API endpoints for processing collections i.e. playlists, albums, shows, and audiobooks"""

    def get_playlist_url(self, playlist: str | Mapping[str, Any] | RemoteResponse, use_cache: bool = True) -> str:
        """
        Determine the type of the given ``playlist`` and return its API URL.
        If type cannot be determined, attempt to find the playlist in the
        list of the currently authorised user's playlists.

        :param playlist: One of the following to identify the playlist URL:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: The playlist URL.
        :raise RemoteIDTypeError: Raised when the function cannot determine the item type of
            the input ``playlist``. Or when it does not recognise the type of the input ``playlist`` parameter.
        """
        if isinstance(playlist, RemoteResponse):
            playlist = playlist.response

        if isinstance(playlist, Mapping):
            if "href" in playlist:
                return playlist["href"]
            elif "id" in playlist:
                return self.wrangler.convert(
                    playlist["id"], kind=RemoteObjectType.PLAYLIST, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL
                )
            elif "uri" in playlist:
                return self.wrangler.convert(
                    playlist["uri"], kind=RemoteObjectType.PLAYLIST, type_in=RemoteIDType.URI, type_out=RemoteIDType.URL
                )

        try:
            return self.wrangler.convert(playlist, kind=RemoteObjectType.PLAYLIST, type_out=RemoteIDType.URL)
        except RemoteIDTypeError:
            playlists = self.get_user_items(kind=RemoteObjectType.PLAYLIST, use_cache=use_cache)
            playlists = {pl["name"]: pl["href"] for pl in playlists}
            if playlist not in playlists:
                raise RemoteIDTypeError(
                    "Given playlist is not a valid URL/URI/ID and name not found in user's playlists",
                    value=playlist
                )
            return playlists[playlist]

    ###########################################################################
    ## POST endpoints
    ###########################################################################
    def create_playlist(self, name: str, public: bool = True, collaborative: bool = False, *_, **__) -> str:
        """
        ``POST: /users/{user_id}/playlists`` - Create an empty playlist for the current user with the given name.

        :param name: Name of playlist to create.
        :param public: Set playlist availability as `public` if True and `private` if False.
        :param collaborative: Set playlist to collaborative i.e. other users may edit the playlist.
        :return: API URL for playlist.
        """
        url = f"{self.url}/users/{self.user_id}/playlists"

        body = {
            "name": name,
            "description": f"Generated using {PROGRAM_NAME}: {PROGRAM_URL}",
            "public": public,
            "collaborative": collaborative,
        }
        pl_url = self.handler.post(url, json=body, log_pad=71)["href"]

        self.logger.debug(f"{'DONE':<7}: {url:<71} | Created playlist: '{name}' -> {pl_url}")
        return pl_url

    def add_to_playlist(
            self,
            playlist: str | Mapping[str, Any] | RemoteResponse,
            items: Collection[str],
            limit: int = 100,
            skip_dupes: bool = True
    ) -> int:
        """
        ``POST: /playlists/{playlist_id}/tracks`` - Add list of tracks to a given playlist.

        :param playlist: One of the following to identify the playlist to add to:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to add.
        :param limit: Size of each batch of IDs to add. This value will be limited to be between ``1`` and ``100``.
        :param skip_dupes: Skip duplicates.
        :return: The number of tracks added to the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent
            a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        url = f"{self.get_playlist_url(playlist, use_cache=False)}/tracks"

        if len(items) == 0:
            self.logger.debug(f"{'SKIP':<7}: {url:<43} | No data given")
            return 0

        self.wrangler.validate_item_type(items, kind=RemoteObjectType.TRACK)

        uri_list = [
            self.wrangler.convert(item, kind=RemoteObjectType.TRACK, type_out=RemoteIDType.URI) for item in items
        ]
        if skip_dupes:  # skip tracks currently in playlist
            pl_current = self.get_items(url, kind=RemoteObjectType.PLAYLIST, use_cache=False)[0]
            tracks_key = self.collection_item_map[RemoteObjectType.PLAYLIST].name.lower() + "s"
            tracks = pl_current[tracks_key][self.items_key]

            uri_current = [track["track"]["uri"] for track in tracks]
            uri_list = [uri for uri in uri_list if uri not in uri_current]

        limit = limit_value(limit, floor=1, ceil=100)
        for uris in batched(uri_list, limit):  # add tracks in batches
            log = [f"Adding {len(uris):>6} items"]
            self.handler.post(url, json={"uris": uris}, log_pad=71, log_extra=log)

        self.logger.debug(f"{'DONE':<7}: {url:<71} | Added {len(uri_list):>6} items to playlist: {url}")
        return len(uri_list)

    ###########################################################################
    ## DELETE endpoints
    ###########################################################################
    def delete_playlist(self, playlist: str | Mapping[str, Any] | RemoteResponse) -> str:
        """
        ``DELETE: /playlists/{playlist_id}/followers`` - Unfollow a given playlist.
        WARNING: This function will destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to delete:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :return: API URL for playlist.
        """
        url = f"{self.get_playlist_url(playlist, use_cache=False)}/followers"
        self.handler.delete(url, log_pad=43)
        return url

    def clear_from_playlist(
            self,
            playlist: str | Mapping[str, Any] | RemoteResponse,
            items: Collection[str] | None = None,
            limit: int = 100
    ) -> int:
        """
        ``DELETE: /playlists/{playlist_id}/tracks`` - Clear tracks from a given playlist.
        WARNING: This function can destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to clear from :
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to remove. If None, clear all songs from the playlist.
        :param limit: Size of each batch of IDs to clear in a single request.
            This value will be limited to be between ``1`` and ``100``.
        :return: The number of tracks cleared from the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent
            a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        url = f"{self.get_playlist_url(playlist, use_cache=False)}/tracks"
        if items is not None and len(items) == 0:
            self.logger.debug(f"{'SKIP':<7}: {url:<43} | No data given")
            return 0

        if items is None:  # clear everything
            pl_current = self.get_items(url, kind=RemoteObjectType.PLAYLIST, extend=True, use_cache=False)[0]

            tracks_key = self.collection_item_map[RemoteObjectType.PLAYLIST].name.lower() + "s"
            tracks = pl_current[tracks_key][self.items_key]
            uri_list = [track[tracks_key.rstrip("s")]["uri"] for track in tracks]
        else:  # clear only the items given
            self.wrangler.validate_item_type(items, kind=RemoteObjectType.TRACK)
            uri_list = [
                self.wrangler.convert(item, kind=RemoteObjectType.TRACK, type_out=RemoteIDType.URI) for item in items
            ]

        if not uri_list:  # skip when nothing to clear
            self.logger.debug(f"{'SKIP':<7}: {url:<43} | No tracks to clear")
            return 0

        limit = limit_value(limit, floor=1, ceil=100)
        for uris in batched(uri_list, limit):  # clear in batches
            body = {"tracks": [{"uri": uri} for uri in uris]}
            log = [f"Clearing {len(uri_list):>3} tracks"]
            self.handler.delete(url, json=body, log_pad=71, log_extra=log)

        self.logger.debug(f"{'DONE':<7}: {url:<71} | Cleared {len(uri_list):>3} tracks")
        return len(uri_list)



"""
Exceptions relating to API operations.
"""
from requests import Response

from musify.exception import MusifyError


class APIError(MusifyError):
    """
    Exception raised for API errors.

    :param message: Explanation of the error.
    :param response: The :py:class:`Response` related to the error.
    """

    def __init__(self, message: str | None = None, response: Response | None = None):
        self.message = message
        self.response = response
        formatted = f"Status code: {response.status_code} | {message}" if response else message
        super().__init__(formatted)


class RequestError(APIError):
    """Exception raised for errors raised when making requests to an API."""



"""
Handle API authorisation for requesting access tokens to an API.
"""
import json
import logging
import os
import socket
from collections.abc import Callable, Mapping, Sequence, MutableMapping
from datetime import datetime
from typing import Any
from urllib.parse import urlparse, parse_qs
from webbrowser import open as webopen

import requests

from musify import PROGRAM_NAME
from musify.api.exception import APIError
from musify.log.logger import MusifyLogger


class APIAuthoriser:
    # noinspection GrazieInspection
    """
    Authorises and validates an API token for given input parameters.
    Functions for returning formatted headers for future, authorised requests.

    Any ``..._args`` parameters must be provided as a dictionary of parameters to be passed directly
    to the :py:class:`requests` module e.g.

    .. code-block:: python

        user_args = {
            "url": token_url,
            "params": {},
            "data": {
                "client_id": client_id,
                "client_secret": client_secret,
                "grant_type": "client_credentials",
            },
            "auth": ('user', 'pass'),
            "headers": {},
            "json": {},
        }

    :param name: The name of the API service being accessed.
    :param auth_args: The parameters to be passed to the
        `requests.post() <https://requests.readthedocs.io/en/latest/api/#requests.post>`_
        function for initial token authorisation. See description for possible example values.
    :param user_args: The parameters to be passed to the
        `requests.post() <https://requests.readthedocs.io/en/latest/api/#requests.post>`_
        function for requesting user authorised access to API services.
        The code response from this request is then added to the authorisation request args
        to grant user authorisation to the API. See description for possible example values.
    :param refresh_args: The parameters to be passed to the
        `requests.post() <https://requests.readthedocs.io/en/latest/api/#requests.post>`_
        function for refreshing an expired token when a refresh_token is present.
        See description for possible example values.
    :param test_args: The parameters to be passed to the
        `requests.get() <https://requests.readthedocs.io/en/latest/api/#requests.get>`_
        function for testing validity of the token. Must be set in conjunction with test_condition to work.
        See description for possible example values.
    :param test_condition: Callable function for testing the response from the given
        test_args. e.g. ``lambda r: "error" not in r``
    :param test_expiry: The time allowance in seconds left until the token is due to expire to use when testing.
        Useful for ensuring the token will be valid for long enough to run your operations e.g.

        - a token has 600 second total expiry time,
        - it is 60 seconds old and therefore still has 540 seconds of authorised time left,
        - you set ``test_expiry`` = 300, the token will pass tests.
        - the same token is tested again later when it is 500 seconds old,
        - it now has only 100 seconds of authorised time left,
        - it will now fail the tests as 100 < 300 and will need to be refreshed.
    :param token: Define a custom input token for initialisation.
    :param token_file_path: Path to use for loading and saving a token.
    :param token_key_path: Keys to the token in auth response. Looks for key 'access_token' by default.
    :param header_key: Header key to apply to headers for authorised calls to the API.
    :param header_prefix: Prefix to add to the header value for authorised calls to the API.
    :param header_extra: Extra data to add to the final headers for future successful requests.
    """

    __slots__ = (
        "logger",
        "name",
        "auth_args",
        "user_args",
        "refresh_args",
        "test_args",
        "test_condition",
        "test_expiry",
        "token",
        "token_file_path",
        "token_key_path",
        "header_key",
        "header_prefix",
        "header_extra",
    )

    _user_auth_socket_address = "localhost"
    _user_auth_socket_port = 8080

    @property
    def token_safe(self) -> dict[str, Any]:
        """Returns a reformatted token, making it safe to log by removing sensitive values at predefined keys."""
        if not self.token:
            return {}
        return {k: f"{v[:5]}..." if str(k).endswith("_token") else v for k, v in self.token.items()}

    @property
    def headers(self) -> dict[str, str]:
        """
        Format headers to usage appropriate format

        :raise APIError: If no token has been loaded,
            or a valid value was not found at the ``token_key_path`` within the token
        """
        if self.token is None:
            raise APIError("Token not loaded.")

        token_value = self.token
        for key in self.token_key_path:  # get token key value at given path
            token_value = token_value.get(key, {})

        if not isinstance(token_value, str):
            raise APIError(
                f"Did not find valid token at key path: {self.token_key_path} -> {token_value} | " +
                str(self.token_safe)
            )

        return {self.header_key: f"{self.header_prefix}{token_value}"} | self.header_extra

    def __init__(
        self,
        name: str,
        auth_args: MutableMapping[str, Any] | None = None,
        user_args: Mapping[str, Any] | None = None,
        refresh_args: Mapping[str, Any] | None = None,
        test_args: Mapping[str, Any] | None = None,
        test_condition: Callable[[str | Mapping[str, Any]], bool] | None = None,
        test_expiry: int = 0,
        token: Mapping[str, Any] | None = None,
        token_file_path: str | None = None,
        token_key_path: Sequence[str] = ("access_token",),
        header_key: str = "Authorization",
        header_prefix: str | None = "Bearer ",
        header_extra: Mapping[str, str] | None = None,
    ):
        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)
        self.name = name

        # maps of requests parameters to be passed to `requests` functions
        self.auth_args: MutableMapping[str, Any] | None = auth_args
        self.user_args: dict[str, Any] | None = user_args
        self.refresh_args: dict[str, Any] | None = refresh_args

        # test params and conditions
        self.test_args: Mapping[str, Any] | None = test_args
        self.test_condition: Callable[[str | Mapping[str, Any]], bool] | None = test_condition
        self.test_expiry: int = test_expiry

        # store token
        self.token: Mapping[str, Any] | None = token
        self.token_file_path: str | None = token_file_path
        self.token_key_path: Sequence[str] = token_key_path

        # information for the final headers
        self.header_key: str = header_key
        self.header_prefix: str = header_prefix or ""
        self.header_extra: dict[str, str] = header_extra or {}

    def load_token(self) -> dict[str, Any] | None:
        """Load stored token from given path"""
        if not self.token_file_path or not os.path.exists(self.token_file_path):
            return self.token

        self.logger.debug("Saved access token found. Loading stored token...")
        with open(self.token_file_path, "r") as file:  # load token
            self.token = json.load(file)
        return self.token

    def save_token(self) -> None:
        """Save new/updated token to given path"""
        if not self.token_file_path or not self.token:
            return

        self.logger.debug(f"Saving token: {self.token_safe}")
        with open(self.token_file_path, "w") as file:
            json.dump(self.token, file, indent=2)

    def authorise(self, force_load: bool = False, force_new: bool = False) -> dict[str, str]:
        """
        Main method for authorisation which tests/refreshes/reauthorises as needed.

        :param force_load: Reloads the token even if it's already been loaded into the object.
            Ignored when force_new is True.
        :param force_new: Ignore saved/loaded token and generate new token.
        :return: Headers for request authorisation.
        :raise APIError: If the token cannot be validated.
        """
        # attempt to load stored token if found
        if force_new:
            self.token = None
        elif self.token is None or force_load:
            self.load_token()

        # generate new token if not or force is enabled
        if self.auth_args and self.token is None:
            log = "Saved access token not found" if self.token is None else "New token generation forced"
            self.logger.debug(f"{log}. Generating new token...")
            self._authorise_user()
            self._request_token(**self.auth_args)

        # test current token
        valid = self.test_token()
        refreshed = False

        # if invalid, first attempt to re-authorise via refresh_token
        if not valid and self.token and "refresh_token" in self.token and self.refresh_args is not None:
            self.logger.debug("Access token is not valid and refresh data found. Refreshing token and testing...")

            if "data" not in self.refresh_args:
                self.refresh_args["data"] = {}
            self.refresh_args["data"]["refresh_token"] = self.token["refresh_token"]
            self._request_token(**self.refresh_args)

            valid = self.test_token()
            refreshed = True

        if not valid and self.auth_args:  # generate new token
            if refreshed:
                log = "Refreshed access token is still not valid"
            else:
                log = "Access token is not valid and and no refresh data found"
            self.logger.debug(f"{log}. Generating new token...")

            self._authorise_user()
            self._request_token(**self.auth_args)
            valid = self.test_token()

        if not self.token:
            raise APIError("Token not generated")
        elif not valid:
            raise APIError(f"Token is still not valid: {self.token_safe}")

        self.logger.debug("Access token is valid. Saving...")
        self.save_token()

        return self.headers

    def _authorise_user(self) -> None:
        """
        Get user authentication code by authorising through user's browser.

        :return: The authentication code
        """
        if not self.user_args or (self.auth_args and self.auth_args.get("data", {}).get("code")):
            return

        self.logger.info_extra("Authorising user privilege access...")

        # set up socket to listen for the redirect from
        socket_listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_listener.bind((self._user_auth_socket_address, self._user_auth_socket_port))
        socket_listener.settimeout(120)
        socket_listener.listen(1)

        print(
            f"\33[1mOpening {self.name} in your browser. "
            f"Log in to {self.name}, authorise, and return here after \33[0m"
        )
        print(f"\33[1mWaiting for code, timeout in {socket_listener.timeout} seconds... \33[0m")

        # add redirect URI to auth_args and user_args
        if not self.auth_args.get("data"):
            self.auth_args["data"] = {}
        if not self.user_args.get("params"):
            self.user_args["params"] = {}
        redirect_uri = f"http://{self._user_auth_socket_address}:{self._user_auth_socket_port}/"
        self.auth_args["data"]["redirect_uri"] = redirect_uri
        self.user_args["params"]["redirect_uri"] = redirect_uri

        # open authorise webpage and wait for the redirect
        auth_response = requests.post(**self.user_args)
        webopen(auth_response.url)
        request, _ = socket_listener.accept()

        request.send(f"Code received! You may now close this window and return to {PROGRAM_NAME}...".encode("utf-8"))
        print("\33[92;1mCode received!\33[0m")
        socket_listener.close()

        # format out the access code from the returned response
        path_raw = next(line for line in request.recv(8196).decode("utf-8").split('\n') if line.startswith("GET"))
        code = parse_qs(urlparse(path_raw).query)["code"][0]

        if "data" not in self.auth_args:
            self.auth_args["data"] = {}
        self.auth_args["data"]["code"] = code

    def _request_token(self, **requests_args) -> dict[str, Any]:
        """
        Authenticates/refreshes basic API access and returns token.

        :param user: Authenticate as the user first to user to generate a user access authenticated token.
        :param data: requests.post() ``data`` parameter to send as a request for authorisation.
        :param requests_args: Other requests.post() parameters to send as a request for authorisation.
        """
        auth_response = requests.post(**requests_args).json()

        # add granted and expiry times to token
        auth_response["granted_at"] = datetime.now().timestamp()
        if "expires_in" in auth_response:
            expires_at = auth_response["granted_at"] + float(auth_response["expires_in"])
            auth_response["expires_at"] = expires_at

        # request sometimes returns new refresh token, append previous one if not
        if "refresh_token" not in auth_response:
            if self.token is not None and "refresh_token" in self.token:
                auth_response["refresh_token"] = self.token["refresh_token"]

        self.token = auth_response
        self.logger.debug(f"New token successfully generated: {self.token_safe}")
        return auth_response

    def test_token(self) -> bool:
        """Test validity of token and given headers. Returns True if all tests pass, False otherwise"""
        if not self.token:
            return False

        self.logger.debug("Begin testing token...")

        token_has_no_error = self._test_no_error()
        if not token_has_no_error:  # skip other tests if error
            return False

        valid_response = self._test_valid_response()
        not_expired = self._test_expiry()

        return token_has_no_error and valid_response and not_expired

    def _test_no_error(self) -> bool:
        """Check if the token contains an error message"""
        result = "error" not in self.token
        self.logger.debug(f"Token contains no error test: {result}")
        return result

    def _test_valid_response(self) -> bool:
        """Check for expected response"""
        if self.test_args is None or self.test_condition is None:
            return True

        response = requests.get(headers=self.headers, **self.test_args)
        try:
            response = response.json()
        except json.JSONDecodeError:
            response = response.text

        result = self.test_condition(response)
        self.logger.debug(f"Valid response test: {result}")
        return result if result is not None else False

    def _test_expiry(self) -> bool:
        """Check if the token is within accepted time range for expiry"""
        if all(key not in self.token for key in ("expires_at", "expires_in")) or self.test_expiry <= 0:
            return True

        if "expires_at" in self.token:
            result = datetime.now().timestamp() + self.test_expiry < self.token["expires_at"]
        else:
            result = self.test_expiry < self.token["expires_in"]

        self.logger.debug(f"Expiry time test: {result}")
        return result



"""
The fundamental core enum classes for the entire package.
"""
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Self

from musify.exception import MusifyEnumError
from musify.types import UnitIterable
from musify.utils import unique_list


class MusifyEnum(IntEnum):
    """Generic class for :py:class:`IntEnum` implementations for the entire package."""

    @classmethod
    def map(cls, enum: Self) -> list[Self]:
        """
        Optional mapper to apply to the enum found during :py:meth:`all`, :py:meth:`from_name`,
        and :py:meth:`from_value` calls
        """
        return [enum]

    @classmethod
    def all(cls) -> list[Self]:
        """Get all enums for this enum."""
        return unique_list(e for enum in cls if enum.name != "ALL" for e in cls.map(enum))

    @classmethod
    def from_name(cls, *names: str, fail_on_many: bool = True) -> list[Self]:
        """
        Returns all enums that match the given enum names

        :param fail_on_many: If more than one enum is found, raise an exception.
        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        names_upper = [name.strip().upper() for name in names]
        enums = unique_list(e for enum in cls if enum.name in names_upper for e in cls.map(enum))

        if len(enums) == 0:
            raise MusifyEnumError(names)
        elif len(enums) > 1 and fail_on_many:
            raise MusifyEnumError(value=enums, message="Too many enums found")

        return enums

    @classmethod
    def from_value(cls, *values: int, fail_on_many: bool = True) -> list[Self]:
        """
        Returns all enums that match the given enum values

        :param fail_on_many: If more than one enum is found, raise an exception.
        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        enums = unique_list(e for enum in cls if enum.value in values for e in cls.map(enum))
        if len(enums) == 0:
            raise MusifyEnumError(values)
        elif len(enums) > 1 and fail_on_many:
            raise MusifyEnumError(value=enums, message="Too many enums found")
        return enums


class Field(MusifyEnum):
    """Base class for field names of an item."""

    @classmethod
    def from_name(cls, *names: str) -> list[Self]:
        """
        Returns all enums that match the given enum names

        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        return super().from_name(*names, fail_on_many=False)

    @classmethod
    def from_value(cls, *values: int) -> list[Self]:
        """
        Returns all enums that match the given enum values

        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        return super().from_value(*values, fail_on_many=False)


class Fields(Field):
    """
    All possible Field enums in this program.

    This is used to ensure all Field enum implementations have the same values for their enum names.
    """
    # all numbers relate to the number of the field value as mapped by MusicBee
    # changing these numbers will break all MusicBee functionality unless it is noted
    # that a mapping with a MusicBee field is not present for that enum

    ALL = 0
    NAME = 1000

    # tags/core properties
    TITLE = 65
    ARTIST = 32
    ALBUM = 30  # MusicBee album ignoring articles like 'the' and 'a' etc.
    ALBUM_ARTIST = 31
    TRACK_NUMBER = 86
    TRACK_TOTAL = 87
    GENRES = 59
    DATE = 900  # no MusicBee mapping
    YEAR = 35
    MONTH = 901  # no MusicBee mapping
    DAY = 902  # no MusicBee mapping
    BPM = 85
    KEY = 903  # unknown MusicBee mapping
    DISC_NUMBER = 52
    DISC_TOTAL = 54
    COMPILATION = 904  # unknown MusicBee mapping
    COMMENTS = 44
    IMAGES = 905  # no MusicBee mapping
    LENGTH = 16
    RATING = 75
    COMPOSER = 43
    CONDUCTOR = 45
    PUBLISHER = 73

    # file properties
    PATH = 106
    FOLDER = 179
    FILENAME = 3
    EXT = 100
    SIZE = 7
    TYPE = 4
    CHANNELS = 8
    BIT_RATE = 10
    BIT_DEPTH = 183
    SAMPLE_RATE = 9

    # date properties
    DATE_CREATED = 921  # no MusicBee mapping
    DATE_MODIFIED = 11
    DATE_ADDED = 12
    LAST_PLAYED = 13

    # miscellaneous properties
    PLAY_COUNT = 14
    DESCRIPTION = 931  # no MusicBee mapping

    # remote properties
    URI = 941  # no MusicBee mapping
    USER_ID = 942  # no MusicBee mapping
    USER_NAME = 943  # no MusicBee mapping
    OWNER_ID = 944  # no MusicBee mapping
    OWNER_NAME = 945  # no MusicBee mapping
    FOLLOWERS = 946  # no MusicBee mapping


@dataclass(frozen=True)
class TagMap:
    """Map of human-friendly tag name to ID3 tag ids for a file type"""
    # helpful for determining tags map: https://wiki.hydrogenaud.io/index.php?title=Tag_Mapping

    title: Sequence[str] = field(default=())
    artist: Sequence[str] = field(default=())
    album: Sequence[str] = field(default=())
    album_artist: Sequence[str] = field(default=())
    track_number: Sequence[str] = field(default=())
    track_total: Sequence[str] = field(default=())
    genres: Sequence[str] = field(default=())
    date: Sequence[str] = field(default=())
    year: Sequence[str] = field(default=())
    month: Sequence[str] = field(default=())
    day: Sequence[str] = field(default=())
    bpm: Sequence[str] = field(default=())
    key: Sequence[str] = field(default=())
    disc_number: Sequence[str] = field(default=())
    disc_total: Sequence[str] = field(default=())
    compilation: Sequence[str] = field(default=())
    comments: Sequence[str] = field(default=())
    images: Sequence[str] = field(default=())

    def __getitem__(self, key: str) -> Sequence[str]:
        """Safely get the value of a given attribute key, returning an empty string if the key is not found"""
        return getattr(self, key, [])


class TagField(Field):
    """Applies extra functionality to :py:class:`Field` for objects which contain modifiable tags"""

    __tags__: frozenset[str] = frozenset(list(TagMap.__annotations__.keys()) + ["uri"])

    @classmethod
    def all(cls, only_tags: bool = False) -> list[Self]:
        """
        Get all enums for this enum.
        When ``only_tags`` is True, returns only those enums that represent a tag for this TagField type.
        """
        enums = super().all()
        if not only_tags:
            return enums

        return list(sorted(cls.from_name(*cls.to_tags(enums)), key=lambda x: enums.index(x)))

    def to_tag(self) -> set[str]:
        """
        Returns all human-friendly tag names for the current enum value.

        Applies mapper to enums before returning as per :py:meth:`map`.
        This will only return tag names if they are found in :py:class:`TagMap`.
        """
        if self == Fields.ALL:
            return {tag.name.lower() for tag in super().all() if tag.name.lower() in self.__tags__}
        return {tag.name.lower() for tag in self.map(self) if tag.name.lower() in self.__tags__}

    @classmethod
    def to_tags(cls, tags: UnitIterable[Self]) -> set[str]:
        """
        Returns all human-friendly tag names for the given enum value.

        Applies mapper to enums before returning as per :py:meth:`map`.
        This will only return tag names if they are found in :py:class:`TagMap`.
        """
        if isinstance(tags, cls):
            return tags.to_tag()
        return {t for tag in tags for t in tag.to_tag()}


class TagFields(TagField):
    """
    All possible TagField enums in this program.

    This is used to ensure all TagField enum implementations have the same values for their enum names.
    """
    ALL = Fields.ALL.value
    NAME = Fields.NAME.value

    # tags/core properties
    TITLE = Fields.TITLE.value
    ARTIST = Fields.ARTIST.value
    ALBUM = Fields.ALBUM.value
    ALBUM_ARTIST = Fields.ALBUM_ARTIST.value
    TRACK_NUMBER = Fields.TRACK_NUMBER.value
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    DATE = Fields.DATE.value
    YEAR = Fields.YEAR.value
    MONTH = Fields.MONTH.value
    DAY = Fields.DAY.value
    BPM = Fields.BPM.value
    KEY = Fields.KEY.value
    DISC_NUMBER = Fields.DISC_NUMBER.value
    DISC_TOTAL = Fields.DISC_TOTAL.value
    COMPILATION = Fields.COMPILATION.value
    COMMENTS = Fields.COMMENTS.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value
    RATING = Fields.RATING.value
    COMPOSER = Fields.COMPOSER.value
    CONDUCTOR = Fields.CONDUCTOR.value
    PUBLISHER = Fields.PUBLISHER.value

    # file properties
    PATH = Fields.PATH.value
    FOLDER = Fields.FOLDER.value
    FILENAME = Fields.FILENAME.value
    EXT = Fields.EXT.value
    SIZE = Fields.SIZE.value
    TYPE = Fields.TYPE.value
    CHANNELS = Fields.CHANNELS.value
    BIT_RATE = Fields.BIT_RATE.value
    BIT_DEPTH = Fields.BIT_DEPTH.value
    SAMPLE_RATE = Fields.SAMPLE_RATE.value

    # date properties
    DATE_CREATED = Fields.DATE_CREATED.value
    DATE_MODIFIED = Fields.DATE_MODIFIED.value
    DATE_ADDED = Fields.DATE_ADDED.value
    LAST_PLAYED = Fields.LAST_PLAYED.value

    # miscellaneous properties
    PLAY_COUNT = Fields.PLAY_COUNT.value
    DESCRIPTION = Fields.DESCRIPTION.value

    # remote properties
    URI = Fields.URI.value
    USER_ID = Fields.USER_ID.value
    USER_NAME = Fields.USER_NAME.value
    OWNER_ID = Fields.OWNER_ID.value
    OWNER_NAME = Fields.OWNER_NAME.value
    FOLLOWERS = Fields.FOLLOWERS.value


ALL_FIELDS = frozenset(Fields.all())
ALL_TAG_FIELDS = frozenset(TagFields.all())



"""
Base functionality to be shared by all classes that implement :py:class:`RemoteAPI` functionality for Spotify.
"""
from abc import ABCMeta
from typing import Any
from urllib.parse import parse_qs, urlparse, urlencode, quote, urlunparse

from musify.libraries.remote.core.api import RemoteAPI


class SpotifyAPIBase(RemoteAPI, metaclass=ABCMeta):
    """Base functionality required for all endpoint functions for the Spotify API"""

    #: The key to reference when extracting items from a collection
    items_key = "items"

    @staticmethod
    def format_next_url(url: str, offset: int = 0, limit: int = 20) -> str:
        """Format a `next` style URL for looping through API pages"""
        url_parsed = urlparse(url)
        params: dict[str, Any] = parse_qs(url_parsed.query)
        params["offset"] = offset
        params["limit"] = limit

        url_parts = list(url_parsed[:])
        url_parts[4] = urlencode(params, doseq=True, quote_via=quote)
        return str(urlunparse(url_parts))



"""
All operations relating to handling of requests to an API.
"""
import json
from collections.abc import Mapping, Iterable
from datetime import datetime, timedelta
from http import HTTPStatus
from time import sleep
from typing import Any

import requests
from requests import Response, Session
from requests_cache import CachedSession, ExpirationTime

from musify.api.authorise import APIAuthoriser
from musify.api.exception import APIError


class RequestHandler(APIAuthoriser):
    """
    Generic API request handler using cached responses for GET requests only.
    Caches GET responses for a maximum of 4 weeks by default.
    Handles error responses and backoff on failed requests.
    See :py:class:`APIAuthoriser` for more info on which params to pass to authorise requests.

    :param cache_path: The path to store the requests session's sqlite cache.
    :param cache_expiry: The expiry time to apply to cached responses after which responses are invalidated.
    :param auth_kwargs: The authorisation kwargs to be passed to :py:class:`APIAuthoriser`.
    """

    __slots__ = ("session", "__dict__")

    #: The initial backoff time for failed requests
    backoff_start = 0.5
    #: The factor by which to increase backoff time for failed requests i.e. backoff_start ** backoff_factor
    backoff_factor = 2
    #: The maximum number of request attempts to make before giving up and raising an exception
    backoff_count = 10

    @property
    def backoff_final(self) -> int:
        """
        The maximum wait time to retry a request in seconds
        until giving up when applying backoff to failed requests
        """
        return self.backoff_start * self.backoff_factor ** self.backoff_count

    @property
    def timeout(self) -> int:
        """
        When the response gives a time to wait until (i.e. retry-after),
        the program will exit if this time is above this timeout (in seconds)
        """
        return sum(self.backoff_start * self.backoff_factor ** i for i in range(self.backoff_count + 1))

    def __init__(self, cache_path: str | None = None, cache_expiry: ExpirationTime = timedelta(weeks=4), **auth_kwargs):
        super().__init__(**auth_kwargs)

        #: The :py:class:`Session` or :py:class:`CachedSession` object
        self.session: CachedSession | Session
        if cache_path:
            self.logger.debug(f"Setting up requests cache: {cache_path}")
            self.session = CachedSession(cache_path, expire_after=cache_expiry, allowable_methods=["GET"])
        else:
            self.session = Session()

    def authorise(self, force_load: bool = False, force_new: bool = False) -> dict[str, str]:
        headers = super().authorise(force_load=force_load, force_new=force_new)
        self.session.headers.update(headers)
        return headers

    def close(self) -> None:
        """Close the current session. No more requests will be possible once this has been called."""
        self.session.close()

    def request(self, method: str, url: str, *args, **kwargs) -> dict[str, Any]:
        """
        Generic method for handling API requests with back-off on failed requests.
        See :py:func:`request` for more arguments.

        :param method: method for the new :class:`Request` object:
            ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        :param url: URL for the new :class:`Request` object.
        :return: The JSON formatted response or, if JSON formatting not possible, the text response.
        :raise APIError: On any logic breaking error/response.
        """
        kwargs.pop("headers", None)
        response = self._request(method=method, url=url, *args, **kwargs)
        backoff = self.backoff_start

        while response is None or response.status_code >= 400:  # error response code received
            waited = False
            if response is not None:
                self._log_response(response=response, method=method, url=url)
                self._handle_unexpected_response(response=response)
                waited = self._handle_wait_time(response=response)

            if not waited and backoff < self.backoff_final:  # exponential backoff
                self.logger.warning(f"Request failed: retrying in {backoff} seconds...")
                sleep(backoff)
                backoff *= self.backoff_factor
            elif not waited:  # max backoff exceeded
                raise APIError("Max retries exceeded")

            response = self._request(method=method, url=url, *args, **kwargs)

        return self._response_as_json(response)

    def _request(
            self,
            method: str,
            url: str,
            use_cache: bool = True,
            log_pad: int = 43,
            log_extra: Iterable[str] = (),
            *args,
            **kwargs
    ) -> Response | None:
        """Handle logging a request, send the request, and return the response"""
        # format logs
        log = [f"{method.upper():<7}: {url:<{log_pad}}"]
        if log_extra:
            log.extend(log_extra)
        if len(args) > 0:
            log.append(f"Args: ({', '.join(args)})")
        if len(kwargs) > 0:
            log.extend(f"{k.title()}: {v}" for k, v in kwargs.items())
        if use_cache and isinstance(self.session, CachedSession):
            if method.upper() in self.session.settings.allowable_methods:
                log.append("Cached")

        headers = self.session.headers
        if "headers" in kwargs:
            headers = kwargs.pop("headers") | self.session.headers

        self.logger.debug(" | ".join(log))
        try:
            if not isinstance(self.session, CachedSession):
                return self.session.request(method=method.upper(), url=url, headers=headers, *args, **kwargs)
            return self.session.request(
                method=method.upper(), force_refresh=not use_cache, url=url, headers=headers, *args, **kwargs
            )
        except requests.exceptions.ConnectionError as ex:
            self.logger.warning(str(ex))
            return

    def _log_response(self, response: Response, method: str, url: str) -> None:
        """Log the method, URL, response text, and response headers."""
        response_headers = response.headers
        if isinstance(response.headers, Mapping):  # format headers if JSON
            response_headers = json.dumps(dict(response.headers), indent=2)
        self.logger.warning(
            f"\33[91m{method.upper():<7}: {url} | Code: {response.status_code} | "
            f"Response text and headers follow:\n"
            f"Response text:\n{response.text}\n"
            f"Headers:\n{response_headers}\33[0m"
        )

    def _handle_unexpected_response(self, response: Response) -> bool:
        """Handle bad response by extracting message and handling status codes that should raise an exception."""
        message = self._response_as_json(response).get("error", {}).get("message")
        error_message_found = message is not None

        if not error_message_found:
            status = HTTPStatus(response.status_code)
            message = f"{status.phrase} | {status.description}"

        if 400 <= response.status_code < 408:
            raise APIError(message, response=response)

        return error_message_found

    def _handle_wait_time(self, response: Response) -> bool:
        """Handle when a wait time is included in the response headers."""
        if "retry-after" not in response.headers:
            return False

        wait_time = int(response.headers["retry-after"])
        wait_str = (datetime.now() + timedelta(seconds=wait_time)).strftime("%Y-%m-%d %H:%M:%S")
        self.logger.info(f"\33[91mRate limit exceeded. Retrying again at {wait_str}\33[0m")

        if wait_time > self.timeout:  # exception if too long
            raise APIError(f"Rate limit exceeded and wait time is greater than timeout of {self.timeout} seconds")

        # wait if time is short
        sleep(wait_time)
        return True

    @staticmethod
    def _response_as_json(response: Response) -> dict[str, Any]:
        """Format the response to JSON and handle any errors"""
        try:
            return response.json()
        except json.decoder.JSONDecodeError:
            return {}

    def get(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends a GET request."""
        kwargs.pop("method", None)
        return self.request("get", url=url, **kwargs)

    def post(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends a POST request."""
        kwargs.pop("method", None)
        return self.request("post", url=url, **kwargs)

    def put(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends a PUT request."""
        kwargs.pop("method", None)
        return self.request("put", url=url, **kwargs)

    def delete(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends a DELETE request."""
        kwargs.pop("method", None)
        return self.request("delete", url, **kwargs)

    def options(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends an OPTIONS request."""
        kwargs.pop("method", None)
        return self.request("options", url=url, **kwargs)

    def head(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends a HEAD request."""
        kwargs.pop("method", None)
        kwargs.setdefault("allow_redirects", False)
        return self.request("head", url=url, **kwargs)

    def patch(self, url: str, **kwargs) -> dict[str, Any]:
        """Sends a PATCH request."""
        kwargs.pop("method", None)
        return self.request("patch", url=url, **kwargs)

    def __copy__(self):
        """Do not copy handler"""
        return self

    def __deepcopy__(self, _: dict = None):
        """Do not copy handler"""
        return self

    def __enter__(self):
        self.authorise()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
All logging filters specific to this package.
"""
import inspect
import logging.handlers
import re
from os.path import join, splitext, split, basename

from musify import PROGRAM_NAME


def format_full_func_name(record: logging.LogRecord, width: int = 40) -> None:
    """
    Set fully qualified path name to function including class name to the given record.
    Optionally, provide a max ``width`` to attempt to truncate the path name to
    by taking only the first letter of each part of the path until the length is equal to ``width``.
    """
    last_call = inspect.stack()[8]

    if record.pathname == __file__:
        # custom logging method has been called, reformat call info to actual call method
        record.pathname = last_call.filename
        record.lineno = last_call.lineno
        record.funcName = last_call.function
        record.filename = basename(record.pathname)
        record.module = record.name.split(".")[-1]

    f_locals = last_call.frame.f_locals
    if "self" not in f_locals:
        path_split = record.name.split(".")
        if record.funcName != "<module>":
            path_split.append(record.funcName)
    else:
        # is a valid and initialised object, extract the class name and determine path to call function from stack
        cls = f_locals["self"].__class__
        path = join(splitext(inspect.getfile(cls))[0], cls.__name__, record.funcName.split(".")[-1])

        folder = ""
        path_split = []
        while not folder.casefold().startswith(PROGRAM_NAME.casefold()) and path:  # get relative path to sources root
            path, folder = split(path)
            path_split.append(folder)
        path_split.append(PROGRAM_NAME.lower())

        # produce fully qualified path
        path_split = list(reversed(path_split[:-1]))

    # truncate long paths by taking first letters of each part until short enough
    path = ".".join(path_split)
    for i, part in enumerate(path_split):
        if len(path) <= width:
            break
        if not part:
            continue

        # take all upper case characters if they exist in part, else, if all lower case, take first letter
        path_split[i] = re.sub("[a-z_]+", "", part) if re.match("[A-Z]", part) else part[0]
        path = ".".join(path_split)

    record.funcName = path


class LogConsoleFilter(logging.Filter):
    """Filter for logging to the console."""

    def __init__(self, name: str = "", module_width: int = 40):
        super().__init__(name)
        self.module_width = module_width

    # noinspection PyMissingOrEmptyDocstring
    def filter(self, record: logging.LogRecord) -> logging.LogRecord | None:
        format_full_func_name(record, width=self.module_width)
        return record


class LogFileFilter(logging.Filter):
    """Filter for logging to a file."""

    def __init__(self, name: str = "", module_width: int = 40):
        super().__init__(name)
        self.module_width = module_width

    # noinspection PyMissingOrEmptyDocstring
    def filter(self, record: logging.LogRecord) -> logging.LogRecord:
        record.msg = re.sub("\33.*?m", "", record.msg)
        format_full_func_name(record, width=self.module_width)
        return record



"""
All logging handlers specific to this package.
"""
import logging.handlers
import os
import shutil
from datetime import datetime
from glob import glob
from os.path import join, dirname, isfile, sep, isdir

from musify.log import LOGGING_DT_FORMAT
from musify.processors.time import TimeMapper


class CurrentTimeRotatingFileHandler(logging.handlers.BaseRotatingHandler):
    """
    Handles log file and directory rotation based on log file/folder name.

    :param filename: The full path to the log file.
        Optionally, include a '{}' part in the path to format in the current datetime.
        When None, defaults to '{}.log'
    :param encoding: When not None, it is used to open the file with that encoding.
    :param when: The timespan for 'interval' which is used together to calculate the timedelta.
        Accepts same values as :py:class:`TimeMapper`.
    :param interval: The multiplier for ``when``.
        When combined with ``when``, gives the negative timedelta relative to now
        which is the maximum datetime to keep logs for.
    :param count: The maximum number of files to keep.
    :param delay: When True, the file opening is deferred until the first call to emit().
    :param errors: Used to determine how encoding errors are handled.
    """
    def __init__(
            self,
            filename: str | None = None,
            encoding: str | None = None,
            when: str | None = None,
            interval: int | None = None,
            count: int | None = None,
            delay: bool = False,
            errors: str | None = None
    ):
        self.dt = datetime.now()

        dt_str = self.dt.strftime(LOGGING_DT_FORMAT)
        if not filename:
            filename = "{}.log"
        filename = filename.replace("\\", sep) if sep == "/" else filename.replace("/", sep)
        self.filename = filename.format(dt_str) if "{}" in filename else filename
        if dirname(self.filename):
            os.makedirs(dirname(self.filename), exist_ok=True)

        self.delta = TimeMapper(when.lower())(interval) if when and interval else None
        self.count = count

        self.removed: list[datetime] = []  # datetime on the files that were removed
        self.rotator(unformatted=filename, formatted=self.filename)

        super().__init__(filename=self.filename, mode="w", encoding=encoding, delay=delay, errors=errors)

    # noinspection PyPep8Naming
    @staticmethod
    def shouldRollover(*_, **__) -> bool:
        """Always returns False. Rotation happens on __init__ and only needs to happen once."""
        return False

    def rotator(self, unformatted: str, formatted: str):
        """
        Rotates the files in the folder on the given ``unformatted`` path.
        Removes files older than ``self.delta`` and the oldest files when number of files >= count
        until number of files <= count. ``formatted`` path is excluded from processing.
        """
        folder = dirname(formatted)
        if not folder:
            return

        # get current files present and prefix+suffix to remove when processing
        paths = tuple(f for f in glob(join(folder, "*")) if f != formatted)
        prefix = unformatted.split("{")[0] if "{" in unformatted and "}" in unformatted else ""
        suffix = unformatted.split("}")[1] if "{" in unformatted and "}" in unformatted else ""

        remaining = len(paths)
        for path in sorted(paths):
            too_many = self.count is not None and remaining >= self.count

            dt_part = path.removeprefix(prefix).removesuffix(suffix)
            try:
                dt_file = datetime.strptime(dt_part, LOGGING_DT_FORMAT)
                too_old = self.delta is not None and dt_file < self.dt - self.delta
            except ValueError:
                dt_file = None
                too_old = False

            is_empty = isdir(path) and not glob(join(path, "*"))

            if too_many or too_old or is_empty:
                os.remove(path) if isfile(path) else shutil.rmtree(path)
                remaining -= 1

                if dt_file and dt_file not in self.removed:
                    self.removed.append(dt_file)



"""
Just the core abstract class for the :py:mod:`Remote` module.
Placed here separately to avoid circular import logic issues.
"""
from abc import abstractmethod, ABCMeta
from typing import Any

from musify.core.base import MusifyObject
from musify.libraries.remote.core.enum import RemoteObjectType


class RemoteResponse(MusifyObject, metaclass=ABCMeta):

    @property
    @abstractmethod
    def response(self) -> dict[str, Any]:
        """The API response for this object"""
        raise NotImplementedError

    @property
    @abstractmethod
    def id(self) -> str:
        """The ID of this item/collection."""
        raise NotImplementedError

    # noinspection PyPropertyDefinition,PyMethodParameters
    @property
    @abstractmethod
    def kind(cls) -> RemoteObjectType:
        """The type of remote object this class represents"""
        raise NotImplementedError

    @abstractmethod
    def refresh(self, skip_checks: bool = False) -> None:
        """
        Refresh this object by updating from the stored API response.
        Useful for updating stored variables after making changes to the stored API response manually.
        """
        raise NotImplementedError



"""
Processor operations that search for and match given items with remote items.

Searches for matches on remote APIs, matches the item to the best matching result from the query,
and assigns the ID of the matched object back to the item.
"""
import logging
from collections.abc import Mapping, Sequence, Iterable, Collection
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import Any

from musify.core.base import MusifyObject, MusifyItemSettable
from musify.core.enum import TagField, TagFields as Tag
from musify.core.result import Result
from musify.exception import MusifyAttributeError
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.enum import RemoteObjectType
from musify.libraries.remote.core.factory import RemoteObjectFactory
from musify.log import REPORT
from musify.log.logger import MusifyLogger
from musify.processors.base import Processor
from musify.processors.match import ItemMatcher
from musify.types import UnitIterable
from musify.utils import align_string, get_max_width


@dataclass(frozen=True)
class ItemSearchResult[T: MusifyItemSettable](Result):
    """Stores the results of the searching process."""
    #: Sequence of Items for which matches were found from the search.
    matched: Sequence[T] = field(default=tuple())
    #: Sequence of Items for which matches were not found from the search.
    unmatched: Sequence[T] = field(default=tuple())
    #: Sequence of Items which were skipped during the search.
    skipped: Sequence[T] = field(default=tuple())


@dataclass(frozen=True)
class SearchConfig:
    """Key settings related to a search algorithm."""
    #: The fields to match results on.
    match_fields: TagField | Iterable[TagField]
    #: A sequence of the tag names to use as search fields in the 1st pass.
    search_fields_1: Sequence[TagField] = (Tag.NAME,)
    #: If no results are found from the tag names in ``search_fields_1`` on the 1st pass,
    #: an optional sequence of the tag names to use as search fields in the 2nd pass.
    search_fields_2: Iterable[TagField] = ()
    #: If no results are found from the tag names in ``search_fields_2`` on the 2nd pass,
    #: an optional sequence of the tag names to use as search fields in the 3rd pass.
    search_fields_3: Iterable[TagField] = ()

    #: The number of the results to request when querying the API.
    result_count: int = 10
    #: The minimum acceptable score for an item to be considered a match.
    min_score: float = 0.1
    #: The maximum score for an item to be considered a perfect match.
    #: After this score is reached by an item, any other items are disregarded as potential matches.
    max_score: float = 0.8
    #: When True, items determined to be karaoke are allowed when matching added items.
    #: Skip karaoke results otherwise.
    allow_karaoke: bool = False


class RemoteItemSearcher(Processor):
    """
    Searches for remote matches for a list of item collections.

    :param matcher: The :py:class:`ItemMatcher` to use when comparing any changes made by the user in remote playlists
        during the checking operation
    :param object_factory: The :py:class:`RemoteObjectFactory` to use when creating new remote objects.
        This must have a :py:class:`RemoteAPI` assigned for this processor to work as expected.
    :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
    """

    __slots__ = ("factory", "use_cache")

    #: The :py:class:`SearchSettings` for each :py:class:`RemoteObjectType`
    search_settings: dict[RemoteObjectType, SearchConfig] = {
        RemoteObjectType.TRACK: SearchConfig(
            match_fields={Tag.TITLE, Tag.ARTIST, Tag.ALBUM, Tag.LENGTH},
            search_fields_1=[Tag.NAME, Tag.ARTIST],
            search_fields_2=[Tag.NAME, Tag.ALBUM],
            search_fields_3=[Tag.NAME],
            result_count=10,
            min_score=0.1,
            max_score=0.8,
            allow_karaoke=False,
        ),
        RemoteObjectType.ALBUM: SearchConfig(
            match_fields={Tag.ARTIST, Tag.ALBUM, Tag.LENGTH},
            search_fields_1=[Tag.NAME, Tag.ARTIST],
            search_fields_2=[Tag.NAME],
            result_count=5,
            min_score=0.1,
            max_score=0.7,
            allow_karaoke=False,
        )
    }

    @property
    def api(self) -> RemoteAPI:
        """The :py:class:`RemoteAPI` to call"""
        return self.factory.api

    def __init__(self, matcher: ItemMatcher, object_factory: RemoteObjectFactory, use_cache: bool = False):
        super().__init__()

        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        #: The :py:class:`ItemMatcher` to use when comparing any changes made by the user in remote playlists
        #: during the checking operation
        self.matcher = matcher
        #: The :py:class:`RemoteObjectFactory` to use when creating new remote objects.
        self.factory = object_factory
        #: When true, use the cache when calling the API endpoint
        self.use_cache = use_cache

    def _get_results(
            self, item: MusifyObject, kind: RemoteObjectType, settings: SearchConfig
    ) -> list[dict[str, Any]] | None:
        """Query the API to get results for the current item based on algorithm settings"""
        self.matcher.clean_tags(item)

        def execute_query(keys: Iterable[TagField]) -> tuple[list[dict[str, Any]], str]:
            """Generate and execute the query against the API for the given item's cleaned ``keys``"""
            attributes = [item.clean_tags.get(key) for key in keys]
            q = " ".join(str(attr) for attr in attributes if attr)
            return self.api.query(q, kind=kind, limit=settings.result_count), q

        results, query = execute_query(settings.search_fields_1)
        if not results and settings.search_fields_2:
            results, query = execute_query(settings.search_fields_2)
        if not results and settings.search_fields_3:
            results, query = execute_query(settings.search_fields_3)

        if results:
            self.matcher.log_messages([item.name, f"Query: {query}", f"{len(results)} results"])
            return results
        self.matcher.log_messages([item.name, f"Query: {query}", "Match failed: No results."], pad="<")

    def _log_results(self, results: Mapping[str, ItemSearchResult]) -> None:
        """Logs the final results of the ItemSearcher"""
        if not results:
            return

        max_width = get_max_width(results)

        total_matched = 0
        total_unmatched = 0
        total_skipped = 0
        total_all = 0

        for name, result in results.items():
            matched = len(result.matched)
            unmatched = len(result.unmatched)
            skipped = len(result.skipped)
            total = total_matched + total_unmatched + total_skipped

            total_matched += matched
            total_unmatched += unmatched
            total_skipped += skipped
            total_all += total

            colour1 = "\33[92m" if matched > 0 else "\33[94m"
            colour2 = "\33[92m" if unmatched == 0 else "\33[91m"
            colour3 = "\33[92m" if skipped == 0 else "\33[93m"

            self.logger.report(
                f"\33[1m{align_string(name, max_width=max_width)} \33[0m|"
                f"{colour1}{matched:>6} matched \33[0m| "
                f"{colour2}{unmatched:>6} unmatched \33[0m| "
                f"{colour3}{skipped:>6} skipped \33[0m| "
                f"\33[97m{total:>6} total \33[0m"
            )

        self.logger.report(
            f"\33[1;96m{'TOTALS':<{max_width}} \33[0m|"
            f"\33[92m{total_matched:>6} matched \33[0m| "
            f"\33[91m{total_unmatched:>6} unmatched \33[0m| "
            f"\33[93m{total_skipped:>6} skipped \33[0m| "
            f"\33[97m{total_all:>6} total \33[0m"
        )
        self.logger.print(REPORT)

    @staticmethod
    def _determine_remote_object_type(obj: MusifyObject) -> RemoteObjectType:
        if hasattr(obj, "kind"):
            return obj.kind
        raise MusifyAttributeError(f"Given object does not specify a RemoteObjectType: {obj.__class__.__name__}")

    def __call__(self, *args, **kwargs) -> dict[str, ItemSearchResult]:
        return self.search(*args, **kwargs)

    def search[T: MusifyItemSettable](
            self, collections: Collection[MusifyCollection[T]]
    ) -> dict[str, ItemSearchResult[T]]:
        """
        Searches for remote matches for the given list of item collections.

        :return: Map of the collection's name to its :py:class:`ItemSearchResult` object.
        """
        self.logger.debug("Searching: START")
        if not [item for c in collections for item in c.items if item.has_uri is None]:
            self.logger.debug("\33[93mNo items to search. \33[0m")
            return {}

        kinds = {coll.__class__.__name__ for coll in collections}
        kind = kinds.pop() if len(kinds) == 1 else "collection"
        self.logger.info(
            f"\33[1;95m ->\33[1;97m "
            f"Searching for matches on {self.api.source} for {len(collections)} {kind}s\33[0m"
        )

        bar = self.logger.get_progress_bar(iterable=collections, desc="Searching", unit=f"{kind}s")
        with ThreadPoolExecutor(thread_name_prefix="searcher-main") as executor:
            search_results = dict(executor.map(lambda coll: (coll.name, self._search_collection(coll)), bar))

        self.logger.print()
        self._log_results(search_results)
        self.logger.debug("Searching: DONE\n")
        return search_results

    def _search_collection[T: MusifyItemSettable](self, collection: MusifyCollection) -> ItemSearchResult[T]:
        kind = collection.__class__.__name__

        skipped = tuple(item for item in collection if item.has_uri is not None)
        if len(skipped) == len(collection):
            self.matcher.log_messages([collection.name, "Skipping search, no items to search"], pad='<')

        if getattr(collection, "compilation", True) is False:
            self.matcher.log_messages([collection.name, "Searching for collection as a unit"], pad='>')
            self._search_collection_unit(collection=collection)

            missing = [item for item in collection.items if item.has_uri is None]
            if missing:
                self.matcher.log_messages(
                    [collection.name, f"Searching for {len(missing)} unmatched items in this {kind}"]
                )
                self._search_items(collection=collection)
        else:
            self.matcher.log_messages([collection.name, "Searching for distinct items in collection"], pad='>')
            self._search_items(collection=collection)

        return ItemSearchResult(
            matched=tuple(item for item in collection if item.has_uri and item not in skipped),
            unmatched=tuple(item for item in collection if item.has_uri is None and item not in skipped),
            skipped=skipped
        )

    def _get_item_match[T: MusifyItemSettable](
            self, item: T, match_on: UnitIterable[TagField] | None = None, results: Iterable[T] = None
    ) -> tuple[T, T | None]:
        kind = self._determine_remote_object_type(item)
        search_config = self.search_settings[kind]

        if results is None:
            responses = self._get_results(item, kind=kind, settings=search_config)
            # noinspection PyTypeChecker
            results: Iterable[T] = map(self.factory[kind], responses or ())

        result = self.matcher(
            item,
            results=results,
            match_on=match_on if match_on is not None else search_config.match_fields,
            min_score=search_config.min_score,
            max_score=search_config.max_score,
            allow_karaoke=search_config.allow_karaoke,
        ) if results else None

        return item, result

    def _search_items[T: MusifyItemSettable](self, collection: Iterable[T]) -> None:
        """Search for matches on individual items in an item collection that have ``None`` on ``has_uri`` attribute"""
        with ThreadPoolExecutor(thread_name_prefix="searcher-items") as executor:
            matches = executor.map(self._get_item_match, filter(lambda i: i.has_uri is None, collection))

        for item, match in matches:
            if match and match.has_uri:
                item.uri = match.uri

    def _search_collection_unit[T: MusifyItemSettable](self, collection: MusifyCollection[T]) -> None:
        """
        Search for matches on an entire collection as a whole
        i.e. search for just the collection and not its distinct items.
        """
        if all(item.has_uri for item in collection):
            return

        kind = self._determine_remote_object_type(collection)
        search_config = self.search_settings[kind]

        responses = self._get_results(collection, kind=kind, settings=search_config)
        key = self.api.collection_item_map[kind]
        for response in responses:
            self.api.extend_items(response, kind=kind, key=key, use_cache=self.use_cache)

        # noinspection PyProtectedMember,PyTypeChecker
        # order to prioritise results that are closer to the item count of the input collection
        results: list[T] = sorted(map(self.factory[kind], responses), key=lambda x: abs(x._total - len(collection)))

        result = self.matcher(
            collection,
            results=results,
            match_on=search_config.match_fields,
            min_score=search_config.min_score,
            max_score=search_config.max_score,
            allow_karaoke=search_config.allow_karaoke,
        )

        if not result:
            return

        with ThreadPoolExecutor(thread_name_prefix="searcher-collection") as executor:
            matches = executor.map(
                lambda item: self._get_item_match(item, match_on=[Tag.TITLE], results=result.items),
                filter(lambda i: i.has_uri is None, collection)
            )

        for item, match in matches:
            if match and match.has_uri:
                item.uri = match.uri

    def as_dict(self) -> dict[str, Any]:
        return {
            "matcher": self.matcher,
            "remote_source": self.factory.api.source,
            "interval": self.use_cache,
        }



"""
Generic processors relating to remote operations.

These can be as simple as a wrangler which converts between ID types for a remote object,
to more complex operations/algorithms such as search+match algorithms.
"""



"""
The framework for interacting with a remote API.

All methods that interact with the API should return raw, unprocessed responses.
"""
import logging
from abc import ABC, abstractmethod
from collections.abc import Collection, MutableMapping, Mapping, Sequence
from typing import Any, Self

from musify.api.request import RequestHandler
from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.enum import RemoteIDType, RemoteObjectType
from musify.libraries.remote.core.processors.wrangle import RemoteDataWrangler
from musify.libraries.remote.core.types import APIInputValue
from musify.log.logger import MusifyLogger
from musify.types import UnitSequence, JSON, UnitList
from musify.utils import align_string, to_collection


class RemoteAPI(ABC):
    """
    Collection of endpoints for a remote API.
    See :py:class:`RequestHandler` and :py:class:`APIAuthoriser`
    for more info on which params to pass to authorise and execute requests.

    :param handler_kwargs: The authorisation kwargs to be passed to :py:class:`APIAuthoriser`.
    """

    __slots__ = ("logger", "handler", "user_data", "wrangler")

    #: Map of :py:class:`RemoteObjectType` for remote collections
    #: to the  :py:class:`RemoteObjectType` of the items they hold
    collection_item_map = {
        RemoteObjectType.PLAYLIST: RemoteObjectType.TRACK,
        RemoteObjectType.ALBUM: RemoteObjectType.TRACK,
        RemoteObjectType.AUDIOBOOK: RemoteObjectType.CHAPTER,
        RemoteObjectType.SHOW: RemoteObjectType.EPISODE,
    }
    #: A set of possible saved item types that can be retrieved for the currently authenticated user
    user_item_types = (
            set(collection_item_map) | {RemoteObjectType.TRACK, RemoteObjectType.ARTIST, RemoteObjectType.EPISODE}
    )
    #: The key to use when getting an ID from a response
    id_key = "id"

    @property
    @abstractmethod
    def user_id(self) -> str | None:
        """ID of the currently authorised user"""
        raise NotImplementedError

    @property
    @abstractmethod
    def user_name(self) -> str | None:
        """Name of the currently authorised user"""
        raise NotImplementedError

    @property
    def url(self) -> str:
        """The base URL of the API"""
        return self.wrangler.url_api

    @property
    def source(self) -> str:
        """The name of the API service"""
        return self.wrangler.source

    def __init__(self, wrangler: RemoteDataWrangler, **handler_kwargs):
        super().__init__()
        #: A :py:class:`RemoteDataWrangler` object for processing URIs
        self.wrangler = wrangler

        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        #: The :py:class:`RequestHandler` for handling authorised requests to the API
        self.handler = RequestHandler(
            name=self.wrangler.source, **{k: v for k, v in handler_kwargs.items() if k != "name"}
        )

        #: Stores the loaded user data for the currently authorised user
        self.user_data: dict[str, Any] = {}

    def authorise(self, force_load: bool = False, force_new: bool = False) -> Self:
        """
        Main method for authorisation, tests/refreshes/reauthorises as needed

        :param force_load: Reloads the token even if it's already been loaded into the object.
            Ignored when force_new is True.
        :param force_new: Ignore saved/loaded token and generate new token.
        :return: Self.
        :raise APIError: If the token cannot be validated.
        """
        self.handler.authorise(force_load=force_load, force_new=force_new)
        return self

    def close(self) -> None:
        """Close the current session. No more requests will be possible once this has been called."""
        self.handler.close()

    ###########################################################################
    ## Misc helpers
    ###########################################################################
    def load_user_data(self) -> None:
        """Load and store user data in this API object for the currently authorised user"""
        self.user_data = self.get_self()

    @staticmethod
    def _merge_results_to_input_mapping(
            original: MutableMapping[str, Any], response: Mapping[str, Any], clear: bool = True,
    ) -> None:
        if clear:
            original.clear()
        original |= response

    def _merge_results_to_input(
            self,
            original: UnitSequence[JSON] | UnitSequence[RemoteResponse],
            responses: UnitList[JSON],
            ordered: bool = True,
            clear: bool = True,
    ) -> None:
        """
        If API response type given on input, update with new results.
        Assumes on a one-to-one relationship between ``original`` and the list of ``results``.

        :param original: The original values given to the function.
        :param responses: The new results from the API.
        :param ordered: When True, function assumes the order of items in ``original`` and ``results`` is the same.
            When False, the function will attempt to match each input value to each result by matching on
            the ``id`` key of each dictionary.
        :param ordered: When True, clear the original value before merging, completely replacing all original data.
        """
        if isinstance(original, str) or not isinstance(original, Collection | RemoteResponse):
            return

        if isinstance(original, RemoteResponse):
            original = [original]
        elif not isinstance(original, Sequence):
            original = to_collection(original)
        if not isinstance(responses, Sequence):
            responses = to_collection(responses, list)

        original = [item.response if isinstance(item, RemoteResponse) else item for item in original]

        valid_types_input = all(isinstance(item, MutableMapping) for item in original)
        valid_types_responses = all(isinstance(item, MutableMapping) for item in responses)
        valid_lengths = len(original) == len(responses)
        if not all((valid_types_input, valid_types_responses, valid_lengths)):
            self.logger.debug(
                "Could not merge responses to given user input | "
                "Reason: assertions failed | "
                f"{valid_types_input=} {valid_types_responses=} {valid_lengths=} | "
            )
            return

        if not ordered:
            expected_keys_input = all(self.id_key in item for item in original)
            expected_keys_responses = all(self.id_key in item for item in responses)
            if not all((expected_keys_input, expected_keys_responses)):
                self.logger.debug(
                    "Could not merge responses to given user input | "
                    f"Reason: unordered and cannot order on {self.id_key=} | "
                    f"{expected_keys_input=} {expected_keys_responses=}"
                )
                return

            id_ordered = [response[self.id_key] for response in original]
            responses.sort(key=lambda response: id_ordered.index(response[self.id_key]))

        for item, response in zip(original, responses):
            self._merge_results_to_input_mapping(original=item, response=response, clear=clear)

    @staticmethod
    def _refresh_responses(responses: Any, skip_checks: bool = False) -> None:
        if isinstance(responses, RemoteResponse):
            responses = [responses]

        for response in responses:
            if isinstance(response, RemoteResponse):
                response.refresh(skip_checks=skip_checks)

    def print_item(
            self, i: int, name: str, uri: str, length: float = 0, total: int = 1, max_width: int = 50
    ) -> None:
        """
        Pretty print item data for displaying to the user.
        Format = ``<i> - <name> | <length> | <URI> - <URL>``.

        :param i: The position of this item in the collection.
        :param name: The name of the item.
        :param uri: The URI of the item.
        :param length: The duration of the item in seconds.
        :param total: The total number of items in the collection
        :param max_width: The maximum width to print names as. Any name lengths longer than this will be truncated.
        """
        print(
            f"\t\33[92m{str(i).zfill(len(str(total)))} \33[0m- "
            f"\33[97m{align_string(name, max_width=max_width)} \33[0m| "
            f"\33[91m{str(int(length // 60)).zfill(2)}:{str(round(length % 60)).zfill(2)} \33[0m| "
            f"\33[93m{uri} \33[0m- "
            f"{self.wrangler.convert(uri, type_in=RemoteIDType.URI, type_out=RemoteIDType.URL_EXT)}"
        )

    @abstractmethod
    def print_collection(
            self,
            value: str | Mapping[str, Any] | RemoteResponse | None = None,
            kind: RemoteIDType | None = None,
            limit: int = 20,
            use_cache: bool = True
    ) -> None:
        """
        Pretty print collection data for displaying to the user.
        Runs :py:meth:`print_item()` for each item in the collection.

        ``value`` may be:
            * A string representing a URL/URI/ID.
            * A remote API JSON response for a collection with a valid ID value under an ``id`` key.
            * A RemoteResponse representing some remote collection of items.

        :param value: The value representing some remote collection. See description for allowed value types.
        :param kind: When an ID is provided, give the kind of ID this is here.
            If None and ID is given, user will be prompted to give the kind anyway.
        :param limit: The number of results to call per request and,
            therefore, the number of items in each printed block.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        """
        raise NotImplementedError

    @abstractmethod
    def get_playlist_url(self, playlist: str | Mapping[str, Any] | RemoteResponse, use_cache: bool = True) -> str:
        """
        Determine the type of the given ``playlist`` and return its API URL.
        If type cannot be determined, attempt to find the playlist in the
        list of the currently authenticated user's playlists.

        :param playlist: In URL/URI/ID form, or the name of one of the currently authenticated user's playlists.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: The playlist URL.
        :raise RemoteIDTypeError: Raised when the function cannot determine the item type of
            the input ``playlist``. Or when it does not recognise the type of the input ``playlist`` parameter.
        """
        raise NotImplementedError

    ###########################################################################
    ## Core - GET endpoints
    ###########################################################################
    @abstractmethod
    def get_self(self, update_user_data: bool = True) -> dict[str, Any]:
        """
        ``GET`` - Get API response for information on current user

        :param update_user_data: When True, update the ``_user_data`` stored in this API object."""
        raise NotImplementedError

    @abstractmethod
    def query(
            self, query: str, kind: RemoteObjectType, limit: int = 10, use_cache: bool = True
    ) -> list[dict[str, Any]]:
        """
        ``GET`` - Query for items. Modify result types returned with kind parameter

        :param query: Search query.
        :param kind: The remote object type to search for.
        :param limit: Number of results to get and return.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: The response from the endpoint.
        """
        raise NotImplementedError

    ###########################################################################
    ## Item - GET endpoints
    ###########################################################################
    @abstractmethod
    def extend_items(
            self,
            response: MutableMapping[str, Any] | RemoteResponse,
            kind: RemoteObjectType | str | None = None,
            key: RemoteObjectType | None = None,
            use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        """
        Extend the items for a given API ``response``.
        The function requests each page of the collection returning a list of all items
        found across all pages for this URL.

        Updates the value of the ``items`` key in-place by extending the value of the ``items`` key with new results.

        If a :py:class:`RemoteResponse`, this function will not refresh itself with the new response.
        The user must call `refresh` manually after execution.

        :param response: A remote API JSON response for an items type endpoint.
        :param kind: The type of response being extended. Optional, used only for logging.
        :param key: The type of response of the child objects.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each item
        """
        raise NotImplementedError

    @abstractmethod
    def get_items(
            self,
            values: APIInputValue,
            kind: RemoteObjectType | None = None,
            limit: int = 50,
            extend: bool = True,
            use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET`` - Get information for given ``values``.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection.
            * A MutableSequence of remote API JSON responses for a collection.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by merging with the new response.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items must all be of the same type of item i.e. all tracks OR all artists etc.
        :param kind: Item type if given string is ID.
        :param limit: When requests can be batched, size of batches to request.
            This value will be limited to be between ``1`` and ``50``.
        :param extend: When True and the given ``kind`` is a collection of items,
            extend the response to include all items in this collection.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each item.
        :raise RemoteObjectTypeError: Raised when the function cannot determine the item type
            of the input ``values``. Or when it does not recognise the type of the input ``values`` parameter.
        """
        raise NotImplementedError

    @abstractmethod
    def get_tracks(
            self, values: APIInputValue, limit: int = 50, use_cache: bool = True, *args, **kwargs,
    ) -> list[dict[str, Any]]:
        """
        Wrapper for :py:meth:`get_items` which only returns Track type responses.
        See :py:meth:`get_items` for more info.
        """
        raise NotImplementedError

    @abstractmethod
    def get_user_items(
            self,
            user: str | None = None,
            kind: RemoteObjectType = RemoteObjectType.PLAYLIST,
            limit: int = 50,
            use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET`` - Get saved items for a given user.

        :param user: The ID of the user to get playlists for. If None, use the currently authenticated user.
        :param kind: Item type to retrieve for the user.
        :param limit: Size of each batch of items to request in the collection items request.
            This value will be limited to be between ``1`` and ``50``.
        :param use_cache: Use the cache when calling the API endpoint. Set as False to refresh the cached response.
        :return: API JSON responses for each collection.
        :raise RemoteIDTypeError: Raised when the input ``user`` does not represent a user URL/URI/ID.
        :raise RemoteObjectTypeError: When the given ``kind`` is not a valid user item/collection.
        """
        raise NotImplementedError

    ###########################################################################
    ## Collection - POST endpoints
    ###########################################################################
    @abstractmethod
    def create_playlist(self, name: str, *args, **kwargs) -> str:
        """
        ``POST`` - Create an empty playlist for the current user with the given name.

        :param name: Name of playlist to create.
        :return: API URL for playlist.
        """
        raise NotImplementedError

    @abstractmethod
    def add_to_playlist(
            self,
            playlist: str | Mapping[str, Any] | RemoteResponse,
            items: Collection[str],
            limit: int = 50,
            skip_dupes: bool = True
    ) -> int:
        """
        ``POST`` - Add list of tracks to a given playlist.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to add.
        :param limit: Size of each batch of IDs to add. This value will be limited to be between ``1`` and ``50``.
        :param skip_dupes: Skip duplicates.
        :return: The number of tracks added to the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent
            a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        raise NotImplementedError

    ###########################################################################
    ## Collection - DELETE endpoints
    ###########################################################################
    @abstractmethod
    def delete_playlist(self, playlist: str | Mapping[str, Any] | RemoteResponse) -> str:
        """
        ``DELETE`` - Unfollow/delete a given playlist.
        WARNING: This function will destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :return: API URL for playlist.
        """
        raise NotImplementedError

    @abstractmethod
    def clear_from_playlist(
            self,
            playlist: str | Mapping[str, Any] | RemoteResponse,
            items: Collection[str] | None = None,
            limit: int = 100
    ) -> int:
        """
        ``DELETE`` - Clear tracks from a given playlist.
        WARNING: This function can destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to clear:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to remove. If None, clear all songs from the playlist.
        :param limit: Size of each batch of IDs to clear in a single request.
            This value will be limited to be between ``1`` and ``100``.
        :return: The number of tracks cleared from the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        raise NotImplementedError

    def __enter__(self):
        self.handler.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handler.__exit__(exc_type, exc_val, exc_tb)



"""
Functionality relating to reading and writing images.
"""
from http.client import HTTPResponse
from io import BytesIO
from pathlib import Path
from urllib.error import URLError
from urllib.request import urlopen, Request

from PIL import Image, UnidentifiedImageError

from musify.file.exception import ImageLoadError


def open_image(source: str | bytes | Path | Request) -> Image.Image:
    """
    Open Image object from a given URL or file path

    :return: The loaded :py:class:`Image.Image`
    :raise ImageLoadError: If the image cannot be loaded.
    """

    try:  # open image from link
        if isinstance(source, Request) or (isinstance(source, str) and source.startswith("http")):
            response: HTTPResponse = urlopen(source)
            image = Image.open(response)
            response.close()
            return image

        elif not isinstance(source, Request):
            return Image.open(source)

    except (URLError, FileNotFoundError, UnidentifiedImageError):
        pass
    raise ImageLoadError(f"{source} | Failed to open image")


def get_image_bytes(image: Image.Image) -> bytes:
    """Extracts bytes from a given Image file"""
    image_bytes_arr = BytesIO()
    image.save(image_bytes_arr, format=image.format)
    return image_bytes_arr.getvalue()
